 /*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * defines.h
 *
 *  Created on: Feb 22, 2011
 *  Author: Timo Minartz
 */

#ifndef DEFINES_H_
#define DEFINES_H_

/*
 * Defines of the ID for the devices: Core/CPU, NIC/Network and Disk
 */

/**
 * ID for all Cores.
 */
#define CORE_ALL_ID 0
/**
 * Valid Core ID's range between #CORE_MIN_ID and #CORE_MAX_ID.
 */
#define CORE_MIN_ID 1
/**
 * Valid Core ID's range between #CORE_MIN_ID and #CORE_MAX_ID.
 */
#define CORE_MAX_ID 99
/**
 * ID for all NICs.
 */
#define NIC_ALL_ID 100
/**
 * Valid Nic ID's range between #NIC_MIN_ID and #NIC_MAX_ID.
 */
#define NIC_MIN_ID 101
/**
 * Valid Nic ID's range between #NIC_MIN_ID and #NIC_MAX_ID.
 */
#define NIC_MAX_ID 199
/**
 * ID for all Disks.
 */
#define DISK_ALL_ID 200
/**
 * Valid Disk ID's range between #DISK_MIN_ID and #DISK_MAX_ID.
 */
#define DISK_MIN_ID 201
/**
 * Valid Disk ID's range between #DISK_MIN_ID and #DISK_MAX_ID.
 */
#define DISK_MAX_ID 299

/**
 * ID for the registered core ids
 */
#define MY_CORE_IDS -1

/**
 * ID for the registered nic ids
 */
#define MY_NIC_IDS -2

/**
 * ID for the registered core ids
 */
#define MY_DISK_IDS -3

/**
 * Rank identifying the initialization through the RMS
 */
#define RMS_RANK -1

/// @cond Doxygen should skip the following code

/**
 * Rounds up to hundreds
 */
#define GET_DEV_TYPE(id) ((id / 100) * 100)

/*
 * Defines to encapsulate the device ID check
 */

#define IS_CORE_ID(id) (id >= CORE_MIN_ID && id <= CORE_MAX_ID)
#define IS_NIC_ID(id) (id >= NIC_MIN_ID && id <= NIC_MAX_ID)
#define IS_DISK_ID(id) (id >= DISK_MIN_ID && id <= DISK_MAX_ID)
#define IS_CORE_ALL(id) (id == CORE_ALL_ID)
#define IS_NIC_ALL(id) (id == NIC_ALL_ID)
#define IS_DISK_ALL(id) (id == DISK_ALL_ID)
#define IS_DEV_ALL(id) (id == CORE_ALL_ID || id == NIC_ALL_ID || id == DISK_ALL_ID)
#define IS_MY_ID(id) (id == MY_CORE_IDS || id == MY_NIC_IDS || id == MY_DISK_IDS)
#define ID_IS_MINE(id, myid) ((myid == MY_CORE_IDS && IS_CORE_ID(id) || myid == MY_NIC_IDS && IS_NIC_ID(id) || myid == MY_DISK_IDS && IS_DISK_ID(id)))

/// @endcond End of skipped code

/* modes for devices, must start at 0 because will be used as indices */

/**
 * Internal Mode, indicating device in an unknown Mode
 */
#define MODE_UNKNOWN 0
/**
 * Mode indicating default (system) mode of devices.
 * For the Cpu, this Mode corresponds to the ondemand governor.
 * For the Nic, this Mode turns on autonegotiation.
 * For the Disk, this Mode remounts disk with system defaults.
 * DO NOT USE THIS MODE - It is only used when finalizing the Daemon.
 */
#define MODE_DEFAULT 1

/**
 * Mode marking a very high utilization for the device - device must be switched to the highest performance mode.
 * For the Cpu, this Mode corresponds to the maximal available frequency (Turboboost frequency if available).
 * For the Nic, this Mode is the same as MODE_MAX.
 * For the Disk, this Mode is the same as MODE_MAX.
 */
#define MODE_TURBO 2

/**
 * Mode marking a high utilization for the device - device must be switched to the high performance mode.
 * For the Cpu, this Mode corresponds to the maximal available frequency (excluding Turboboost frequency).
 * For the Nic, this Mode corresponds to maximal speed and full duplex.
 * For the Disk, this Mode wakes the Disk if in MODE_MIN or MODE_UNUSED.
 */
#define MODE_MAX 3

/**
 * Mode marking a mid-range utilization for the device - if possible, device can be switched to a mid-range performance mode.
 * For the Cpu, this Mode corresponds to the mid-range available frequency.
 * For the Nic, this Mode corresponds to mid-range speed and full duplex.
 * For the Disk, this Mode is the same as MODE_MAX.
 */
#define MODE_MED 4

/**
 * Mode marking a low utilization for the device - if possible, device can be switched to a low performance mode.
 * For the Cpu, this Mode corresponds to the minimal available frequency.
 * For the Nic, this Mode corresponds to minimal speed and full duplex.
 * For the Disk, this Mode corresponds to hdparm -y, which force the IDE drive to immediately enter the low power consumption standby mode, usually causing it to spin down.
 */
#define MODE_MIN 5

/**
 * Mode marking the device as unused - which means a device can possibly be switched to sleep.
 * For the CPU, this is Mode is the same as MODE_MIN.
 * For the NIC, this Mode corresponds to minimal speed and half duplex.
 * For the Disk, this Mode corresponds to hdparm -Y, which forces the IDE drive to immediately enter the lowest power consumption sleep mode.
 */
#define MODE_UNUSED 6

/// @cond EXCLUDE_IN_API

#define COUNT_MODES 7

typedef int Mode;

typedef int Type;

#define TYPE_TIME_REQUEST 0
#define TYPE_MODE_REQUEST 1

static char* MODE_STRINGS[COUNT_MODES] = {"MODE_UNKNOWN","MODE_DEFAULT","MODE_TURBO", "MODE_MAX", "MODE_MED", "MODE_MIN", "MODE_UNUSED"};

#define MODE_TO_STRING(mode) (MODE_STRINGS[mode])

/* states for devices */

#define STATE_UNKNOWN -1
#define STATE_SWITCHING 0
#define STATE_STANDBY 2
#define STATE_IDLE 3
#define STATE_ACTIVE_IDLE 4
#define STATE_SPINDOWN 5
#define STATE_SPINUP 6

typedef int State;

/* governors for the cpus */

#define GOV_USERSPACE "userspace"
#define GOV_ONDEMAND "ondemand"
#define GOV_PERFORMANCE "performance"
#define GOV_POWERSAVE "powersave"
#define GOV_UNKNOWN "unknown"

typedef char* Governor;

/* frequencies for the cpus */

#define FREQ_UNKNOWN 1

typedef int Frequency;

/* path the default config file */
#define DEFAULT_CONFIG_FILE "/etc/eedaemon.conf"

/* error messages */

#define ERROR_CPU_READ_NUM_CORES 2
#define ERROR_CORE_READ_FREQUENCIES 3
#define CPU_NOT_INITIALIZED 4
#define SOCKET_NOT_INITIALIZED 5
#define CORE_NOT_INITIALIZED 6
#define NIC_NOT_INITIALIZED 7

#define ERROR_CONNECTION_FAILED 8
#define ERROR_STREAM_CREATION 9
#define ERROR_THREAD_CREATION 10

/// @endcond EXCLUDE_IN_API

#endif /* DEFINES_H_ */
