/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Command.c
 *
 *  Created on: Apr 16, 2011
 *  Author: Timo Minartz
 */

#include "Command.h"
#include "print.h"

#include <string.h>
#include <stdio.h>

#define check_read_all_return_value(ret) \
	if (ret == FALSE && error->code == G_IO_ERROR_CANCELLED) {\
		g_error_free(error);\
		return FALSE;\
	} else if(ret == FALSE) {\
		error("Error read data from stream: %s\n",error->message);\
		g_error_free(error);\
		return FALSE;\
	} else if(count != bytes_read) {\
		error("Error reading bytes from stream: Read %lu out of %lu bytes.\n", bytes_read, count);\
		return FALSE;\
	}

#define check_write_all_return_value(ret) \
	if (ret == FALSE) {\
		error("Error write data to stream: %s\n",error->message);\
		g_error_free(error);\
		return FALSE;\
	}

Command* command_new(const gchar* format, ...) {

	va_list params;

	va_start (params, format);

	Command *command = g_malloc(sizeof(Command));
	command->variant = g_variant_new_va(format, NULL, &params);

	va_end (params);

	return command;
}

gchar* command_get_name(Command *command) {

	gchar* name = NULL;

	if (command != NULL && command->variant != NULL) {

		GVariantIter iter;
		g_variant_iter_init(&iter, command->variant);

		GVariant *child = g_variant_iter_next_value(&iter);
		g_variant_get(child, g_variant_get_type_string(child), &name);

		g_variant_unref(child);
	}

	return name;
}

void command_unref(Command *command) {
	if (command != NULL) {
		g_variant_unref(command->variant);
		g_free(command);
	}
}

gboolean command_send(GOutputStream *stream, Command *command) {

	if (command == NULL || stream == NULL)
		return FALSE;

	gsize size = g_variant_get_size(command->variant);

	gpointer bytes = g_malloc(size);

	g_variant_store(command->variant, bytes);

	GError *error = NULL;
	gboolean ret = g_output_stream_write_all(stream, &size, sizeof(gsize),
			NULL, NULL, &error);

	check_write_all_return_value(ret);

	const gchar * type_string = (gchar*) g_variant_get_type_string(
			command->variant);

	gsize len = strlen(type_string);

	ret = g_output_stream_write_all(stream, &len, sizeof(gsize), NULL, NULL,
			&error);

	check_write_all_return_value(ret);

	ret = g_output_stream_write_all(stream, type_string, len, NULL, NULL,
			&error);

	check_write_all_return_value(ret);

	ret = g_output_stream_write_all(stream, bytes, size, NULL, NULL, &error);

	check_write_all_return_value(ret);

	g_free(bytes);

	return TRUE;
}

void destroy(gpointer data) {
	g_free((gchar*) data);
}

gboolean command_receive(GInputStream *stream, GCancellable *cancellable,
		Command **command) {

	/** TODO: if receive failed (unknown bytes etc., then initiate termination of connection */

	if (command == NULL || stream == NULL)
		return FALSE;

	gsize size;
	gsize bytes_read = 0;
	gsize count = 0;

	GError *error = NULL;

	count = sizeof(gsize);
	gboolean ret = g_input_stream_read_all(stream, &size, count, &bytes_read,
			cancellable, &error);

	check_read_all_return_value(ret);

	gsize len = 0;

	ret = g_input_stream_read_all(stream, &len, count, &bytes_read,
			cancellable, &error);

	check_read_all_return_value(ret);

	gchar type_string[len + 1];

	count = len;
	ret = g_input_stream_read_all(stream, type_string, count, &bytes_read, cancellable,
			&error);

	check_read_all_return_value(ret);

	type_string[len] = '\0';

	if(!g_variant_type_string_is_valid(type_string)) {
		// may be not valid if client applications crashes...
		error("No valid data in stream: \"%s\". Maybe connection terminated unexpected.\n", type_string);
		return FALSE;
	}

	GVariantType *type = g_variant_type_new(type_string);

	gpointer bytes = g_malloc(size);

	count = size;
	ret = g_input_stream_read_all(stream, bytes, count, &bytes_read, cancellable,
			&error);

	check_read_all_return_value(ret);

	GVariant *variant = g_variant_new_from_data(type, bytes, size, TRUE,
			&destroy, bytes);

	check_read_all_return_value(ret);

	*command = g_malloc(sizeof(Command));
	(*command)->variant = variant;

	g_variant_type_free(type);

	return TRUE;

}

gboolean command_execute(Command *command, GModule *module) {

	if (command == NULL || module == NULL)
		return FALSE;

	/*
	 * Implement manual type string parsing for all known types from client_eed.c:
	 *
	 * eed_send_command(command_new("(siisi)", "ee_init", tag, rank, prgname, core_id));
	 * eed_send_command(command_new("(siiii)", "ee_dev_mode", device_id, mode_id, tag, rank));
	 * eed_send_command(command_new("(siiiii)", "ee_dev_mode_in", device_id, mode_id, secs, tag, rank));
	 * eed_send_command(command_new("(sii)", "ee_finalize", tag, rank));
	 * eed_send_command(command_new("(si)", "ee_finalize_all", tag_to_finalize));
	 */

	typedef void (* func_si)(gint);
	typedef void (* func_sii)(gint, gint);
	typedef void (* func_siiii)(gint, gint, gint, gint);
	typedef void (* func_siisi)(gint, gint, gchar *, gint);
	typedef void (* func_siiiii)(gint, gint, gint, gint, gint);

	gchar* name;

	/*
	 * This string belongs to GVariant and must not be freed.
	 */
	gchar * type_string = (gchar*) g_variant_get_type_string(command->variant);

	if (strcmp(type_string, "(si)") == 0) {
		func_si func;
		gint param1;
		g_variant_get(command->variant, type_string, &name, &param1);
		if (!g_module_symbol(module, name, (gpointer *) &func)) {
			//g_free(name);
			return FALSE;
		}
		func(param1);
		g_free(name);
		return TRUE;

	} else if (strcmp(type_string, "(sii)") == 0) {
		func_sii func;
		gint param1;
		gint param2;
		g_variant_get(command->variant, type_string, &name, &param1, &param2);

		if (!g_module_symbol(module, name, (gpointer *) &func)) {
			//g_free(name);
			return FALSE;
		}
		func(param1, param2);
		g_free(name);
		return TRUE;

	} else if (strcmp(type_string, "(siisi)") == 0) {
		func_siisi func;
		gint param1;
		gint param2;
		gchar * param3;
		gint param4;
		g_variant_get(command->variant, type_string, &name, &param1, &param2,
				&param3, &param4);

		if (!g_module_symbol(module, name, (gpointer *) &func)) {
			//g_free(name);
			return FALSE;
		}
		func(param1, param2, param3,param4);
		g_free(name);
		g_free(param3);
		return TRUE;

	} else if (strcmp(type_string, "(siiii)") == 0) {
		func_siiii func;
		gint param1;
		gint param2;
		gint param3;
		gint param4;
		g_variant_get(command->variant, type_string, &name, &param1, &param2,
				&param3, &param4);
		if (!g_module_symbol(module, name, (gpointer *) &func)) {
			//g_free(name);
			return FALSE;
		}
		func(param1, param2, param3, param4);
		g_free(name);
		return TRUE;
	} else if (strcmp(type_string, "(siiiii)") == 0) {
		func_siiiii func;
		gint param1;
		gint param2;
		gint param3;
		gint param4;
		gint param5;
		g_variant_get(command->variant, type_string, &name, &param1, &param2,
				&param3, &param4, &param5);
		if (!g_module_symbol(module, name, (gpointer *) &func)) {
			//g_free(name);
			return FALSE;
		}
		func(param1, param2, param3, param4, param5);
		g_free(name);
		return TRUE;
	}

	return FALSE;

}

gchar* command_get_string(Command *command) {

	if (command != NULL) {

		/*
		 * Implement manual type string parsing for all known types (see command_execute())
		 */

		gchar* name;

		gchar * type_string = (gchar*) g_variant_get_type_string(
				command->variant);

		if (strcmp(type_string, "(si)") == 0) {
			gint param1;
			g_variant_get(command->variant, type_string, &name, &param1);
			gchar * string = g_malloc(strlen(name) + 10);
			g_snprintf(string, strlen(name) + 9, "%s(%d)", name, param1);
			g_free(name);
			return string;

		} else if (strcmp(type_string, "(sii)") == 0) {
			gint param1;
			gint param2;
			g_variant_get(command->variant, type_string, &name, &param1,
					&param2);
			gchar * string = g_malloc(strlen(name) + 20);
			g_snprintf(string, strlen(name) + 19, "%s(%d,%d)", name, param1,
					param2);
			g_free(name);
			return string;

		} else if (strcmp(type_string, "(siisi)") == 0) {
			gint param1;
			gint param2;
			gchar * param3;
			gint param4;
			g_variant_get(command->variant, type_string, &name, &param1,
					&param2, &param3, &param4);
			gchar * string = g_malloc(strlen(name) + 30 + strlen(param3));
			g_snprintf(string, strlen(name) + 29 + strlen(param3), "%s(%d,%d,%s,%d)", name, param1,
					param2, param3,param4);
			g_free(name);
			g_free(param3);
			return string;

		} else if (strcmp(type_string, "(siiii)") == 0) {
			gint param1;
			gint param2;
			gint param3;
			gint param4;
			g_variant_get(command->variant, type_string, &name, &param1,
					&param2, &param3, &param4);
			gchar * string = g_malloc(strlen(name) + 40);
			g_snprintf(string, strlen(name) + 39, "%s(%d,%d,%d,%d)", name,
					param1, param2, param3, param4);
			g_free(name);
			return string;
		} else if (strcmp(type_string, "(siiiii)") == 0) {
			gint param1;
			gint param2;
			gint param3;
			gint param4;
			gint param5;
			g_variant_get(command->variant, type_string, &name, &param1,
					&param2, &param3, &param4, &param5);
			gchar * string = g_malloc(strlen(name) + 40);
			g_snprintf(string, strlen(name) + 39, "%s(%d,%d,%d,%d,%d)", name,
					param1, param2, param3, param4, param5);
			g_free(name);
			return string;
		}

	}

	return NULL;
}

void command_print(Command * command) {
	gchar * string = command_get_string(command);

	if (string == NULL)
		g_print("(null)");
	else
		g_print("%s", string);

	g_free(string);

}
