/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * client_eed.c
 *
 *  Created on: Dec 6, 2010
 *      Author: Timo Minartz
 */

#include <glib.h>
#include <stdlib.h> /* for getenv */
#include <stdio.h> /* for sscanf */
#include <string.h> /* for strdup */
#include <sched.h> /* for sched_getcpu */

#include "eed.h"
#include "Client.h"
#include "print.h"

static gint rank = 0; /* rank of this client instance, will be set in ee_init */
static gint tag = 0; /* tag of this client instance, will be set in ee_init */

static Client * client = NULL; /* Client instance to communicate commands to the server Daemon */
static gchar * name = NULL;

void ee_init(gint * argc, gchar *** argv, gint new_tag, gint new_rank) {

	if (client == NULL) {

		/* initialize thread safe glib */
		if (!g_thread_get_initialized())
			g_thread_init(NULL);

#ifdef G_THREADS_ENABLED
		registerThread(g_thread_self(), "eed");
#endif

		// export EED_VERBOSITY=1 to get verbosity...
		gchar *verbstr = getenv("EED_VERBOSITY");
		if (verbstr != NULL)
			sscanf(verbstr, "%d", &verbose);
		else
			verbose = FALSE;

#ifdef DEBUG
		// export EED_DEBUG=1/2/3 to get debugmessages
		verbstr = getenv("EED_DEBUG");
		if (verbstr != NULL) {
			sscanf(verbstr, "%d", &print_debug_messages);
			if(print_debug_messages < 0) {
				print_debug_messages = 0;
				print("Debugging disabled via environment variable.\n");
			} else if (print_debug_messages > 3){
				print_debug_messages = 3;
			}
		} else {
			print_debug_messages = 1;
		}
#else
		print_debug_messages = 0; /* do not print debug messages */
#endif

		info("Be verbose.\n");
		debug(1, "Debuglevel: %d.\n", print_debug_messages);

		client = client_new(1313);

		gchar ** entries = g_strsplit((*argv)[0], "/", 20);

		name = g_strdup(entries[g_strv_length(entries) - 1]);

		g_strfreev(entries);

	}

	if (client != NULL) {

		// send tag, rank and core_id to the daemon -> rank for unique identification, tag for tracing and job unregistering, core_id for resource mapping

		tag = new_tag;
		rank = new_rank;

		gint core_id = -1;

		core_id = sched_getcpu();

		if (core_id >= 0) {

			client_send_command(client,
					command_new("(siisi)", "ee_init", tag, rank, name,
							core_id));

			info(
					"Initialized client library for application %s with tag %d and rank %d on core %d\n.", name, tag, rank, core_id);
		} else {
			error(
					"Getting cpu id failed while instantiating client for application %s.", name);
		}
	} else {
		error("Instantiation of Client failed for application %s.", name);
	}

}

void ee_init_rms(gint * argc, gchar *** argv, gint new_rank) {
	gint jobid = -1;
	gint new_tag = 0;

	// at the moment, only pbs as rms is implemented
	gchar *verbstr = getenv("PBS_JOBID");
	if (verbstr != NULL)
		sscanf(verbstr, "%d", &jobid);

	if (jobid == -1) {
		error(
				"Cannot read environment variable PBS_JOBID. Using default tag 0.\n");
	} else {
		new_tag = jobid;
	}

	ee_init(argc, argv, new_tag, new_rank);
}

void ee_init_rms_all(gint * argc, gchar *** argv) {
	gint jobid = -1;
	gint new_tag = 0;

	// at the moment, only pbs as rms is implemented
	gchar *verbstr = getenv("PBS_JOBID");
	if (verbstr != NULL)
		sscanf(verbstr, "%d", &jobid);

	if (jobid == -1) {
		error(
				"Cannot read environment variable PBS_JOBID. Using default tag %d.\n", new_tag);
	} else {
		new_tag = jobid;
	}

	ee_init(argc, argv, new_tag, RMS_RANK);
}

void ee_dev_mode(int device_id, int mode_id) {

	if (client != NULL) {
		client_send_command(client,
				command_new("(siiii)", "ee_dev_mode", device_id, mode_id, tag,
						rank));

		debug(1, "Send command ee_dev_mode(%d,%d)\n", device_id, mode_id);
	}
}

void ee_dev_mode_in(int device_id, int mode_id, int secs) {

	if (client != NULL) {
		client_send_command(client,
				command_new("(siiiii)", "ee_dev_mode_in", device_id, mode_id,
						secs, tag, rank));

		debug(1,
				"Send command ee_dev_mode_in(%d,%d,%d)\n", device_id, mode_id, secs);
	}
}

void ee_finalize() {

	if (client != NULL) {
		client_send_command(client,
				command_new("(sii)", "ee_finalize", tag, rank));

		client_unref(client);
		client = NULL;
		g_free(name);
		name = NULL;

		info("Closed connection for tag %d with rank %d\n.", tag, rank);
	}
}

void ee_finalize_all(int tag_to_finalize) {

	if (client == NULL) {
		/* initialize thread safe glib */
		if (!g_thread_get_initialized())
			g_thread_init(NULL);

		client = client_new(1313);
	}

	client_send_command(client,
			command_new("(si)", "ee_finalize_all", tag_to_finalize));
	client_unref(client);
	client = NULL;

	if (name != NULL)
		g_free(name);

	info("Closed connection for all with tag %d\n.", tag_to_finalize);
}
