/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Config.c
 *
 *  Created on: Jul 5, 2011
 *      Author: Timo Minartz
 */

#include <stdlib.h> /* for atoi*/
#include <string.h> /* for strlen, strcat */
#include <stdio.h> /* for sscanf */

#include "Config.h"
#include "print.h"
#include "defines.h"
#include "Core.h"
#include "Disk.h"
#include "Nic.h"
#include "Resource.h"
#include "ConstrainedResource.h"

#define MAX_DEV_LENGTH 20
#define MAX_LINE_LENGTH 100

gint set_total_num_cores(gint *total_num_cores) {

	gchar * buffer;
	gsize length;
	GError *error = NULL;

	if (g_file_get_contents("/sys/devices/system/cpu/present", &buffer,
			&length, &error))
		*total_num_cores = atoi(buffer + 2) + 1;
	else {
		debug(1, "Error reading from cpu file: %s\n", error->message);
		g_error_free(error);
		return ERROR_CPU_READ_NUM_CORES;
	}

	g_free(buffer);

	return 0;
}

gboolean config_write(Config *config, gchar * filename) {

	GKeyFile *keyfile;
	GKeyFileFlags flags;
	GError *error = NULL;
	gsize length;

	/* Create a new GKeyFile object and a bitwise list of flags. */
	keyfile = g_key_file_new();
	flags = G_KEY_FILE_KEEP_COMMENTS;

	/* Processor group */

	gint cores_per_socket = g_hash_table_size(config->socketMapping)
			/ config->countSockets;
	gint cores[config->countSockets][cores_per_socket];

	for (gint i = 0; i < g_hash_table_size(config->socketMapping); ++i) {
		gint j = GPOINTER_TO_INT(
				g_hash_table_lookup(config->socketMapping, GINT_TO_POINTER(i)));
		cores[j][i % cores_per_socket] = i;
	}

	for (gint i = 0; i < config->countSockets; ++i) {
		GString * key = g_string_new("Socket");
		g_string_append_printf(key, "%d", i);
		g_key_file_set_integer_list(keyfile, "Processor", key->str, cores[i],
				cores_per_socket);
		GString * comment = g_string_new(" List with cores located on ");
		g_string_append_printf(comment, "%s (e.g. 0;[1;...])", key->str);
		g_key_file_set_comment(keyfile, "Processor", key->str, comment->str, NULL);
		g_string_free(key, TRUE);
		g_string_free(comment, TRUE);
	}

	g_key_file_set_comment(keyfile, "Processor", NULL, " Processor group, each line represents a core list for a socket", NULL);

	/* Devices group */

	// nic devices, separated by comma
	const gchar* devices[g_hash_table_size(config->deviceMapping)];
	gint idx = 0;
	for (gint i = NIC_MIN_ID; i <= NIC_MAX_ID; ++i) {
		gpointer value = g_hash_table_lookup(config->deviceMapping,
				GINT_TO_POINTER(i));
		if (value == NULL)
			break;

		devices[idx++] = (gchar *) value;
	}

	g_key_file_set_string_list(keyfile, "Devices", "NIC", devices, idx);

	g_key_file_set_comment(keyfile, "Devices", "NIC", " List with all network devices (e.g. eth0;[eth1;...])", NULL);

	idx = 0;
	for (gint i = DISK_MIN_ID; i <= DISK_MAX_ID; ++i) {
		gpointer value = g_hash_table_lookup(config->deviceMapping,
				GINT_TO_POINTER(i));
		if (value == NULL)
			break;

		devices[idx++] = (gchar *) value;
	}

	g_key_file_set_string_list(keyfile, "Devices", "Disk", devices, idx);

	g_key_file_set_comment(keyfile, "Devices", "Disk", " List with all harddrives (e.g. /dev/sda;[/dev/sdb;...])", NULL);

	g_key_file_set_comment(keyfile, "Devices", NULL, " Devices group, each line represents a device list for each type (NIC and Disk)", NULL);

	/* Settings group */

	g_key_file_set_integer(keyfile, "Settings", "Processor State Change",
			config->startup_times[0]);

	g_key_file_set_comment(keyfile, "Settings", "Processor State Change", " Time in microseconds needed for a core to be in MODE_MAX (e.g. 100)", NULL);

	g_key_file_set_integer(keyfile, "Settings", "NIC State Change",
			config->startup_times[1]);

	g_key_file_set_comment(keyfile, "Settings", "NIC State Change", " Time in microseconds needed for a network device to be in MODE_MAX (e.g. 100)", NULL);

	g_key_file_set_integer(keyfile, "Settings", "Disk State Change",
			config->startup_times[2]);

	g_key_file_set_comment(keyfile, "Settings", "Disk State Change", " Time in microseconds needed for a harddisk to be in MODE_MAX (e.g. 100)", NULL);

	g_key_file_set_boolean(keyfile, "Settings",
			"Core Shared Voltage Regulator", !config->frequency_per_core);

	g_key_file_set_comment(keyfile, "Settings", "Core Shared Voltage Regulator", " If all cores on a socket can only have the same frequency (true/false)", NULL);

	g_key_file_set_integer(keyfile, "Settings", "Resource Update Interval",
			config->interval);

	g_key_file_set_comment(keyfile, "Settings", "Resource Update Interval", " Interval in microseconds for mode updates of the request per device (e.g. 100)", NULL);

	g_key_file_set_comment(keyfile, "Settings", NULL, " Settings for devices, needed for decision process", NULL);

	gchar * content = g_key_file_to_data(keyfile, &length, NULL);

	g_key_file_free(keyfile);

	if (content == NULL) {
		return FALSE;
	}

	gboolean ret = g_file_set_contents(filename, content, -1, &error);

	g_free(content);

	return ret;
}

gboolean config_read(Config **config, gchar * filename) {

	if ((*config) != NULL) {
		error("Config must be NULL!\n");
		return FALSE;
	}

	GKeyFile *keyfile;
	GKeyFileFlags flags;
	GError *error = NULL;
	gsize length;

	/* Create a new GKeyFile object and a bitwise list of flags. */
	keyfile = g_key_file_new();
	flags = G_KEY_FILE_KEEP_COMMENTS;

	/* Load the GKeyFile from keyfile.conf or return. */
	if (!g_key_file_load_from_file(keyfile, filename, flags, &error)) {
		error("Error loading config file: %s\n", error->message);
		return FALSE;
	}

	GHashTable * socketMapping =
			g_hash_table_new(g_direct_hash, g_direct_equal);
	GList *disks = NULL;
	GList *nics = NULL;
	gint frequency_per_core;
	glong interval;

	/* Read in data from the key file from the group "Processor". */

	//	countSockets = g_key_file_get_integer(keyfile, "Processor",
	//			"Count Sockets", NULL);

	gint socketID = 0;
	while (TRUE) {
		GString * key = g_string_new("Socket");
		g_string_append_printf(key, "%d", socketID);
		gint * cores = g_key_file_get_integer_list(keyfile, "Processor",
				key->str, &length, NULL);

		g_string_free(key, TRUE);

		if(cores == NULL || length == 0)
			break;

		for (gint i = 0; i < length; ++i) {
			g_hash_table_insert(socketMapping, GINT_TO_POINTER(cores[i]),
					GINT_TO_POINTER(socketID));
			debug(1, "Added core %d -> socket %d\n", cores[i], socketID);
		}

		g_free(cores);

		socketID++;
	}

	gint countSockets = socketID;

	gchar ** devices = g_key_file_get_string_list(keyfile, "Devices", "NIC",
			&length, NULL);

	for (gint i = 0; i < g_strv_length(devices); ++i) {
		if (strcmp(g_strstrip(devices[i]), "") != 0) {
			nics = g_list_prepend(nics, g_strdup(g_strstrip(devices[i])));
			debug(1, "Added nic: %s\n", devices[i]);
		}
	}

	g_strfreev(devices);

	devices = g_key_file_get_string_list(keyfile, "Devices", "Disk", &length,
			NULL);

	for (gint i = 0; i < g_strv_length(devices); ++i) {
		if (strcmp(g_strstrip(devices[i]), "") != 0) {
			disks = g_list_prepend(disks, g_strdup(g_strstrip(devices[i])));
			debug(1, "Added disk: %s\n", devices[i]);
		}
	}

	g_strfreev(devices);

	// startup times, separated by comma
	glong * startup_times = g_malloc(3 * sizeof(glong));

	startup_times[0] = g_key_file_get_integer(keyfile, "Settings",
			"Processor State Change", NULL);
	startup_times[1] = g_key_file_get_integer(keyfile, "Settings",
			"NIC State Change", NULL);
	startup_times[2] = g_key_file_get_integer(keyfile, "Settings",
			"Disk State Change", NULL);

	frequency_per_core = !g_key_file_get_boolean(keyfile, "Settings",
			"Core Shared Voltage Regulator", NULL);

	interval = g_key_file_get_integer(keyfile, "Settings",
			"Resource Update Interval", NULL);

	(*config) = config_new(countSockets, socketMapping, disks, nics,
			startup_times, frequency_per_core, interval);

	//free lists
	for (GList* it = g_list_first(nics); it != NULL; it = it->next) {
		g_free(it->data);
	}

	g_list_free(nics);

	for (GList* it = g_list_first(disks); it != NULL; it = it->next) {
		g_free(it->data);
	}

	g_list_free(disks);

	g_key_file_free(keyfile);

	if (config == NULL)
		return FALSE;

	return TRUE;
}

Config * config_new(gint countSockets, GHashTable * socketMapping,
		GList* disks, GList* nics, glong* startup_times,
		gboolean frequency_per_core, glong interval) {

	Config * config = g_malloc(sizeof(Config));

	config->frequency_per_core = frequency_per_core;
	config->countSockets = countSockets;
	config->socketMapping = socketMapping;
	config->interval = interval;
	config->startup_times = startup_times;

	gint count_cores = g_hash_table_size(socketMapping);

	/*** device mapping ***/

	// Cores
	config->deviceMapping = g_hash_table_new(g_direct_hash, g_direct_equal);

	for (gint i = 0; i < count_cores; ++i) {
		g_hash_table_insert(config->deviceMapping, GINT_TO_POINTER(i + 1),
		// at the moment, ignore the startup time...
				GINT_TO_POINTER(i)); // Core 0,1,2...
	}

	// Nics
	gint i = 1;
	for (GList* it = g_list_first(nics); it != NULL; it = it->next) {
		g_hash_table_insert(config->deviceMapping, GINT_TO_POINTER(100 + i),
				g_strdup(it->data));
		i++;
	}

	// Disks
	i = 1;
	for (GList* it = g_list_first(disks); it != NULL; it = it->next) {
		g_hash_table_insert(config->deviceMapping, GINT_TO_POINTER(200 + i),
				g_strdup(it->data));
		i++;
	}

	return config;
}

Config * config_create_AMD() {

	GList *disks = NULL;
	disks = g_list_append(disks, "/dev/sda");
	GList * nics = NULL;
	nics = g_list_append(nics, "eth0");

	// static socket mapping

	GHashTable * socketMapping =
			g_hash_table_new(g_direct_hash, g_direct_equal);

	/*
	 * -------------------------------------------------------------
	 * Socket 0: ( 0 1 2 3 4 5 6 7 8 9 10 11 )
	 * Socket 1: ( 12 13 14 15 16 17 18 19 20 21 22 23 )
	 * -------------------------------------------------------------
	 */

	for (gint i = 0; i < 12; i++) {
		// socket 0
		g_hash_table_insert(socketMapping, GINT_TO_POINTER(i),
				GINT_TO_POINTER(0));

		// socket 1
		g_hash_table_insert(socketMapping, GINT_TO_POINTER(i + 12),
				GINT_TO_POINTER(1));
	}

	glong * startup_times = g_malloc(3 * sizeof(glong));

	startup_times[0] = 200;
	startup_times[1] = 200;
	startup_times[2] = 200;

	Config * config = config_new(2, socketMapping, disks, nics, startup_times,
			TRUE, 200);

	//free lists
	g_list_free(nics);
	g_list_free(disks);

	return config;
}

Config * config_create_Intel() {

	GList *disks = NULL;
	disks = g_list_append(disks, "/dev/sda");
	GList * nics = NULL;
	nics = g_list_append(nics, "eth0");

	// static socket mapping

	GHashTable * socketMapping =
			g_hash_table_new(g_direct_hash, g_direct_equal);

	gint num_cores = 1;

	set_total_num_cores(&num_cores);

	if (num_cores == 16) {

		/* SMT enabled */

		/*
		 * -------------------------------------------------------------
		 * Socket 0: ( 0 8 1 9 2 10 3 11 )
		 * Socket 1: ( 4 12 5 13 6 14 7 15 )
		 * -------------------------------------------------------------
		 */

		for (gint i = 0; i < 4; i++) {
			// socket 0
			g_hash_table_insert(socketMapping, GINT_TO_POINTER(i),
					GINT_TO_POINTER(0));

			// socket 1
			g_hash_table_insert(socketMapping, GINT_TO_POINTER(i + 4),
					GINT_TO_POINTER(1));
		}

		for (gint i = 8; i < 12; i++) {
			// socket 0
			g_hash_table_insert(socketMapping, GINT_TO_POINTER(i),
					GINT_TO_POINTER(0));

			// socket 1
			g_hash_table_insert(socketMapping, GINT_TO_POINTER(i + 4),
					GINT_TO_POINTER(1));
		}

	} else if (num_cores == 8) {

		/* SMT disabled */

		/*
		 * -------------------------------------------------------------
		 * Socket 0: ( 0 1 2 3 )
		 * Socket 1: ( 4 5 6 7 )
		 * -------------------------------------------------------------
		 */

		for (gint i = 0; i < 4; i++) {
			// socket 0
			g_hash_table_insert(socketMapping, GINT_TO_POINTER(i),
					GINT_TO_POINTER(0));

			// socket 1
			g_hash_table_insert(socketMapping, GINT_TO_POINTER(i + 4),
					GINT_TO_POINTER(1));
		}

	} else {
		error("Unknown num_cores: %d\n", num_cores);
		return NULL;
	}

	glong * startup_times = g_malloc(3 * sizeof(glong));

	startup_times[0] = 200;
	startup_times[1] = 200;
	startup_times[2] = 200;

	Config * config = config_new(2, socketMapping, disks, nics, startup_times,
			FALSE, 200);

	//free lists
	g_list_free(nics);
	g_list_free(disks);

	return config;
}

Config * config_create_testConfig() {

	GList *disks = NULL;
	disks = g_list_append(disks, "/dev/sda");
	GList * nics = NULL;
	nics = g_list_append(nics, "eth0");

	// static socket mapping

	GHashTable * socketMapping =
			g_hash_table_new(g_direct_hash, g_direct_equal);

	gint num_cores = 8;

	/*
	 * -------------------------------------------------------------
	 * Socket 0: ( 0 1 2 3 )
	 * Socket 1: ( 4 5 6 7 )
	 * -------------------------------------------------------------
	 */

	for (gint i = 0; i < 4; i++) {
		// socket 0
		g_hash_table_insert(socketMapping, GINT_TO_POINTER(i),
				GINT_TO_POINTER(0));

		// socket 1
		g_hash_table_insert(socketMapping, GINT_TO_POINTER(i + 4),
				GINT_TO_POINTER(1));
	}

	glong * startup_times = g_malloc(3 * sizeof(glong));

	startup_times[0] = 200;
	startup_times[1] = 200;
	startup_times[2] = 200;

	Config * config = config_new(2, socketMapping, disks, nics, startup_times,
			FALSE, 200);

	//free lists
	g_list_free(nics);
	g_list_free(disks);

	return config;
}

gint config_get_global_core_id(Config * config, gint local_core_id,
		gint socket_id) {

	// count cores per socket is the total number of cores divided by the socket count
	gint count_cores_per_socket = g_hash_table_size(config->socketMapping)
			/ config->countSockets;

	return count_cores_per_socket * socket_id + local_core_id;

}

ResourcesTable * config_create_resources_table(Config * config) {

	ResourcesTable * rt = resources_table_new();

	ConstrainedResource * crs[config->countSockets];

	for (gint i = 0; i < config->countSockets; ++i) {
		crs[i] = NULL;
	}

	gint total_count_cores = 0;

	// create all devices and resources

	GHashTableIter iter;
	gpointer key, value;

	g_hash_table_iter_init(&iter, config->deviceMapping);
	while (g_hash_table_iter_next(&iter, &key, &value)) {

		gint id = GPOINTER_TO_INT(key);

		if (IS_CORE_ID(id)) {

			gint core_id = GPOINTER_TO_INT(value);

			Core * core = core_new(core_id);

			if (core != NULL) {

				// create new resource for this core
				Resource * resource = resource_new(id, core);

				ConstrainedResource * cr = NULL;

				if (config->frequency_per_core) {
					cr = constrained_resource_new(config->interval);
					resources_table_add_constrained_resource(rt, cr);

				} else {

					gpointer key2, value2;

					if (g_hash_table_lookup_extended(config->socketMapping,
							GINT_TO_POINTER(core_id), &key2, &value2)) {

						gint socketID = GPOINTER_TO_INT(value2);

						if (crs[socketID] == NULL) {
							crs[socketID] = constrained_resource_new(
									config->interval);
							resources_table_add_constrained_resource(rt,
									crs[socketID]);
						}

						cr = crs[socketID];
					}

				}

				constrained_resource_register(cr, resource);

				resources_table_add(rt, resource);

			} else {
				error(
						"Core %d could not be allocated - Maybe no frequency scaling available?\n", core_id);
				error("Further calls to cpufreq will be ignored.\n");
			}

		} else if (IS_NIC_ID(id)) {

			gchar * device = (gchar*) value;
			Nic * nic = nic_new_dev(device, config->startup_times[2]);

			if (nic != NULL) {

				Resource * resource = resource_new(id, nic);

				ConstrainedResource * cr = constrained_resource_new(
						config->interval);

				resources_table_add_constrained_resource(rt, cr);

				constrained_resource_register(cr, resource);

				resources_table_add(rt, resource);

			} else {
				error(
						"Nic could not be allocated on device %s - Maybe not able to create socket?\n", device);
				error("Further calls to ethtool will be ignored.\n");
			}

		} else if (IS_DISK_ID(id)) {

			gchar * device = (gchar*) value;
			Disk * disk = disk_new_dev(device, config->startup_times[1]);

			if (disk != NULL) {

				Resource * resource = resource_new(id, disk);

				ConstrainedResource * cr = constrained_resource_new(
						config->interval);

				resources_table_add_constrained_resource(rt, cr);

				constrained_resource_register(cr, resource);

				resources_table_add(rt, resource);

			} else {
				error(
						"Disk could not be allocated on device %s - Maybe not sufficient privileges?\n", device);
				error("Further calls to hdparm will be ignored.\n");
			}

		} else {
			// unknown device id...
			error(
					"Unknown device id %d: Device id must be >= 0 and < 30\n", id);
		}
	}

	return rt;
}

void config_unref(Config * config) {

	g_hash_table_unref(config->socketMapping);

	//free disk devices
	for (gint i = DISK_MIN_ID; i <= DISK_MAX_ID; ++i) {
		gpointer value = g_hash_table_lookup(config->deviceMapping,
				GINT_TO_POINTER(i));

		if (value == NULL)
			break;

		g_free(value);
	}

	// free nic devices
	for (gint i = NIC_MIN_ID; i <= NIC_MAX_ID; ++i) {
		gpointer value = g_hash_table_lookup(config->deviceMapping,
				GINT_TO_POINTER(i));
		if (value == NULL)
			break;

		g_free(value);
	}

	g_hash_table_unref(config->deviceMapping);
	g_free(config->startup_times);
	g_free(config);
}

