/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Config.h
 *
 *  Created on: Jul 5, 2011
 *      Author: Timo Minartz
 */

#ifndef CONFIG_H_
#define CONFIG_H_

#include <glib.h>
#include "ResourcesTable.h"
#include "defines.h"


/**
 * Hardware specific Config struct containing
 * - the mapping of the Cores to the Sockets
 * - if frequency scaling is possible per Core or per Socket
 * - the devices and corresponding device IDs
 * - interval in microseconds for mode updates of the request per device
 * - the count of sockets
 */
typedef struct {
	/**
	 * (Key,Value) = (CoreID,SocketID)
	 * To find fast the correponding to socket for a provided core
	 */
	GHashTable * socketMapping;
	GHashTable * deviceMapping;
	/**
	 * Time in microseconds needed for each device to be in MODE_MAX
	 */
	glong * startup_times;
	gboolean frequency_per_core;
	glong interval;
	gint countSockets;
} Config;

/**
 * Creates a new Config for the eeClust AMD nodes based on static hardware information.
 */
Config * config_create_AMD();

/**
 * Creates a new Config for the eeClust Intel nodes based on static hardware information.
 */
Config * config_create_Intel();

Config * config_create_testConfig();

/**
 * Creates a new Config
 * @param countSockets count of CPU sockets for node
 * @param socketMapping Mapping of cores to sockets. Expects a Map <core_number,socket_number>.
 * @param disks Lists of device names for Disks (sda,sdb2,hda,...)
 * @param nics Lists of device names for Nics (eth0,eth1,...)
 * @param startup_times Array of size 3 for startup times for the devices (Core, Disk, Nic). The startup time for each type is assumed to be the same.
 * @param frequency_per_core If the frequency is adjustable per core or per socket
 * @param interval Interval in milliseconds for updating device modes (e.g. 200)
 * @return a new allocated Config
 */
Config * config_new(gint countSockets, GHashTable * socketMapping, GList* disks, GList* nics, glong* startup_times, gboolean frequency_per_core, glong interval);

/**
 * Writes a Config to the file specified
 * @param config Config to write
 * @param path_to_file Path to file
 * @return TRUE if file is writable and data is written to file
 */
gboolean config_write(Config *config, gchar * path_to_file);

/**
 * Reads a Config from to the file specified
 * @param config Pointer to new Config, must be NULL
 * @param path_to_file Path to file
 * @return TRUE if file is readable, data makes sense and config is NULL
 */
gboolean config_read(Config **config, gchar * filename);

/**
 * Creates a GHashTable containing all devices which created based on the Config.
 * That means all Cores, Nics and Disks are instantiated based on the ID in the Config.
 * Further the Socket objects are created to manage the associated Cores.
 * @param config Config to create devices from
 * @return GHashTable with deviceID as key and the device itself as value
 */
GHashTable * config_create_devices(Config * config);

ResourcesTable * config_create_resources_table(Config * config);

/**
 * Frees a Config object.
 * @param config Config to free
 */
void config_unref(Config * config);

#endif /* CONFIG_H_ */
