/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * ConfigGenerator.c
 *
 *  Created on: October 12, 2011
 *      Author: Timo Minartz
 */

#include <unistd.h> /* for get_current_dir_name */
#include <string.h> /* for strlen, strstr */
#include <stdlib.h> /* for atoi */
#include <sys/ioctl.h> /* for ioctl */
#include <netinet/in.h> /* for ifru */
#include <linux/if.h> /* for ifreq */

#include "Config.h"
#include "Core.h"
#include "Nic.h"
#include "Disk.h"

#ifdef HAVE_DBC
#include "DBConnector.h"
#include "MetricInterpreter.h"
#endif
#define MAX_WAIT_TIME 20 // in seconds
GList* get_disk_device_names() {

	GList * devices = NULL;
	gchar * last = NULL;

	gchar * command[] = { "ls", "/sys/block", NULL };
	gchar * standard_out = NULL;
	gchar * standard_err = NULL;
	gint exit_status = 0;
	GError* error = NULL;

	gchar * dir_name = get_current_dir_name();

	gboolean ret = g_spawn_sync(dir_name, command, NULL, G_SPAWN_SEARCH_PATH,
			NULL, NULL, &standard_out, &standard_err, &exit_status, &error);

	g_free(dir_name);

	if (!ret) {
		g_print("%s\n", error->message);
		g_error_free(error);
		return devices;
	}

	if (standard_out != NULL) {

		gchar ** lines = g_strsplit(standard_out, "\n", 100);

		gint i = 0;
		gint count = g_strv_length(lines);

		while (i < count - 1) {
			// e.g. loop0 or sda
			gchar * device = lines[i];
			if (last == NULL || !g_str_has_prefix(device, last)) {

				// only if device does not end with a digit
				if (device[strlen(device) - 1] < 48 || device[strlen(device)
						- 1] > 57) {
					last = device;
					devices = g_list_append(devices, g_strconcat("/dev/", device, NULL));
				}
			}
			i++;
		}

		g_strfreev(lines);

	} else {
		g_print("Executing command failed: %s\n", standard_err);
		return devices;
	}

	g_print("Found %d disk(s): ", g_list_length(devices));

	for (GList* it = g_list_first(devices); it != NULL; it = it->next)
		g_print("%s ", (gchar *) it->data);

	g_print("\n");

	g_free(standard_err);
	g_free(standard_out);

	return devices;
}

GList* get_nic_device_names() {

	GList * devices = NULL;

	char buf[1024];
	struct ifconf ifc;
	struct ifreq *ifr;
	int sck;
	int nInterfaces;
	int i;

	/* Get a socket handle. */
	sck = socket(AF_INET, SOCK_DGRAM, 0);
	if (sck < 0) {
		g_print("Cannot create socket\n");
		return devices;
	}

	/* Query available interfaces. */
	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = buf;
	if (ioctl(sck, SIOCGIFCONF, &ifc) < 0) {
		g_print("Cannot execute ioctl(SIOCGIFCONF)\n");
		return devices;
	}

	/* Iterate through the list of interfaces. */
	ifr = ifc.ifc_req;
	nInterfaces = ifc.ifc_len / sizeof(struct ifreq);
	for (i = 0; i < nInterfaces; i++) {
		struct ifreq *item = &ifr[i];

		/* ignore loopback and virtual devices */
		if (!g_str_has_prefix(item->ifr_name, "lo") && !g_str_has_prefix(
				item->ifr_name, "vir")) {
			devices = g_list_append(devices, g_strdup(item->ifr_name));
		}
	}

	g_print("Found %d network device(s): ", g_list_length(devices));

	for (GList* it = g_list_first(devices); it != NULL; it = it->next)
		g_print("%s ", (gchar *) it->data);

	g_print("\n");

	return devices;
}

GHashTable* get_core_to_socket_mapping(gint *count_sockets) {

	// figure out which cores are where in the system.

	gchar * command[] = { "likwid-topology", NULL };
	gchar * standard_out = 0;
	gint exit_status = 0;
	GError* error = NULL;

	gchar * dir_name = get_current_dir_name();

	gboolean ret = g_spawn_sync(dir_name, command, NULL,
			G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_SEARCH_PATH, NULL, NULL,
			&standard_out, NULL, &exit_status, &error);

	g_free(dir_name);

	if (!ret || g_str_has_prefix(standard_out, "Please load module ")) {
		if (error != NULL) {
			g_print("%s\nModule for likwid-topology loaded?\n", error->message);
			g_error_free(error);
		} else {
			g_print("Please load module for likwid-topology.\n");
		}
		return NULL;
	}

	if (standard_out != NULL) {

		GHashTable * socketMapping = g_hash_table_new(g_direct_hash,
				g_direct_equal);

		gchar ** lines = g_strsplit(standard_out, "\n", 100);

		gint i = 0;
		gint count = g_strv_length(lines);

		while (i < count && !g_str_has_prefix(lines[i], "Sockets:"))
			i++;

		// line: Sockets:		2

		gchar ** line = g_strsplit(lines[i], ":", 2);
		*count_sockets = atoi(line[1]);
		g_strfreev(line);

		i++;

		// line: Cores per socket:	4

		line = g_strsplit(lines[i], ":", 2);
		gint cores_per_socket = atoi(line[1]);
		g_strfreev(line);

		i++;

		// line: Threads per core:	2

		line = g_strsplit(lines[i], ":", 2);
		gint threads_per_core = atoi(line[1]);
		g_strfreev(line);

		while (i < count && !g_str_has_prefix(lines[i], "Socket 0:"))
			i++;

		for (int j = 0; j < *count_sockets; ++j) {

			// line: Socket 0: ( 0 8 1 9 2 10 3 11 )

			g_print("%s\n", lines[i]);

			gchar * ids = strstr(lines[i], "(");
			ids++; // skip bracket
			g_strstrip(ids);
			ids[strlen(ids) - 1] = '\0';
			g_strstrip(ids);

			// ids: 0 8 1 9 2 10 3 11
			gchar ** idvector = g_strsplit(ids, " ",
					cores_per_socket * threads_per_core);

			for (int k = 0; k < cores_per_socket * threads_per_core; ++k) {
				g_hash_table_insert(socketMapping,
						GINT_TO_POINTER(atoi(idvector[k])), GINT_TO_POINTER(j));
				g_print("Core %d is mapped to socket %d\n", atoi(idvector[k]),
						j);
			}

			g_strfreev(idvector);
			i++;

		}

		g_strfreev(lines);
		g_free(standard_out);

		return socketMapping;

	}

	g_print("Executing command failed: No output\n");
	return NULL;

}

glong getTimestamp() {
	GTimeVal gTimeVal;
	g_get_current_time(&gTimeVal);

	glong microseconds = gTimeVal.tv_sec * 1000000 + gTimeVal.tv_usec;
	return microseconds;
}

glong * get_startup_times(GList * disks, GList * nics) {
	glong * startup_times = g_malloc(3 * sizeof(long));

	// we can assume at least one core ;)
	Core * core = core_new(0);

	// DEFAULT -> MAX
	glong before = getTimestamp();
	core_switch_to_mode(core, MODE_MAX);

	if (core->has_turboboost)
		while (core_get_state(core) != core->frequencies[1]) {
		}
	else
		while (core_get_state(core) != core->frequencies[0]) {
		}

	glong after = getTimestamp();

	g_print(
			"Core: DEFAULT -> MAX (includes initial governor change): %ld microseconds\n",
			after - before);

	// MAX -> MIN

	before = getTimestamp();
	core_switch_to_mode(core, MODE_UNUSED);

	while (core_get_state(core) != core->frequencies[core->count_frequencies
			- 1]) {
	}

	after = getTimestamp();

	g_print("Core: MAX -> UNUSED: %ld microseconds\n", after - before);

	// MIN -> MAX

	before = getTimestamp();
	core_switch_to_mode(core, MODE_MAX);

	if (core->has_turboboost)
		while (core_get_state(core) != core->frequencies[1]) {
		}
	else
		while (core_get_state(core) != core->frequencies[0]) {
		}

	after = getTimestamp();

	g_print("Core: UNUSED -> MAX: %ld microseconds\n", after - before);

	startup_times[0] = after - before;

	Nic * nic = nic_new_dev((gchar *) g_list_first(nics)->data, 200);

	// DEFAULT -> MAX

	before = getTimestamp();
	nic_switch_to_mode(nic, MODE_MAX);

	while (nic_get_state(nic) == STATE_SWITCHING) {
	}

	after = getTimestamp();

	g_print("Nic: DEFAULT -> MAX: %ld microseconds\n", after - before);

	// MAX -> MIN

	before = getTimestamp();
	nic_switch_to_mode(nic, MODE_UNUSED);

	while (nic_get_state(nic) == STATE_SWITCHING) {
	}

	after = getTimestamp();

	g_print("Nic: MAX -> UNUSED: %ld microseconds\n", after - before);

	// MIN -> MAX

	before = getTimestamp();
	nic_switch_to_mode(nic, MODE_MAX);

	while (nic_get_state(nic) == STATE_SWITCHING) {
	}

	after = getTimestamp();

	g_print("Nic: UNUSED -> MAX: %ld microseconds\n", after - before);

	startup_times[1] = after - before;

	Disk * disk = disk_new_dev((gchar *) g_list_first(disks)->data, 200);

	// DEFAULT -> MAX

	before = getTimestamp();
	disk_switch_to_mode(disk, MODE_MAX);

	gint counter = 0;
	while (counter < MAX_WAIT_TIME && disk_get_state(disk) != STATE_ACTIVE_IDLE) {
		sleep(1);
		counter++;
	}

	after = getTimestamp();

	if (counter < MAX_WAIT_TIME)
		g_print("Disk: DEFAULT -> MAX: %ld microseconds\n", after - before);
	else
		g_print("Disk: DEFAULT -> MAX: > %d seconds\n", MAX_WAIT_TIME);

	// MAX -> MIN

	before = getTimestamp();
	disk_switch_to_mode(disk, MODE_MIN);

	counter = 0;
	while (counter < MAX_WAIT_TIME && disk_get_state(disk) != STATE_STANDBY) {
		sleep(1);
		counter++;
	}

	after = getTimestamp();

	if (counter < MAX_WAIT_TIME)
		g_print("Disk: MAX -> UNUSED: %ld microseconds\n", after - before);
	else
		g_print("Disk: MAX -> USED: > %d seconds\n", MAX_WAIT_TIME);

	// UNUSED -> MAX

	before = getTimestamp();
	disk_switch_to_mode(disk, MODE_MAX);

	counter = 0;
	while (counter < MAX_WAIT_TIME && disk_get_state(disk) != STATE_ACTIVE_IDLE) {
		sleep(1);
		counter++;
	}

	after = getTimestamp();

	g_print("Disk: UNUSED -> MAX: %ld microseconds\n", after - before);

	if (counter < MAX_WAIT_TIME) {
		g_print("Disk: UNUSED -> MAX: %ld microseconds\n", after - before);
		startup_times[2] = after - before;
	} else {
		g_print(
				"Disk: UNUSED -> MAX: > %d seconds, assuming %d seconds for startup time.\n",
				MAX_WAIT_TIME, MAX_WAIT_TIME);
		startup_times[2] = MAX_WAIT_TIME * 1000000; // seconds -> microseconds...
	}

	disk_unref(disk);

	nic_unref(nic);

	core_unref(core);

	return startup_times;
}

gboolean get_frequency_per_core(gint count_sockets, GHashTable * socketMapping) {

#ifdef HAVE_DBC

	// TODO: create load here... otherwise measurements do not make sense at all!

	gchar host[10];
	gethostname(host, 10);

	gint count_cores = g_hash_table_size(socketMapping);

	// create cores ordered by sockets
	Core * sockets[count_sockets][count_cores / count_sockets];
	int indices[count_sockets];

	for(gint i=0; i<count_sockets; ++i) {
		indices[i] = 0;
	}

	for (gint i = 0; i < count_cores; ++i) {

		gint socketNo = GPOINTER_TO_INT(g_hash_table_lookup(socketMapping, GINT_TO_POINTER(i)));

		sockets[socketNo][indices[socketNo]] = core_new(i);
		indices[socketNo] = indices[socketNo] + 1;
	}

	// print out mapping
	for(gint i=0; i<count_sockets; ++i) {
		g_print("Socket %d: (", i);
		for(gint j=0; j<count_cores / count_sockets; ++j) {
			g_print(" %d ", ((Core *) sockets[i][j])->id);
		}
		g_print(")\n");
	}

	// all cores in MODE_DEFAULT => ondemand governor

	glong before = getTimestamp();
	sleep(5);
	glong after = getTimestamp();

	row_struct * p_nodeRow;

	gint crows = db_getNodeMetric("power", host, before, after, &p_nodeRow);

	gdouble power_ondemand = mi_getMeanPowerConsumption(&p_nodeRow, crows);

	free(p_nodeRow);

	for(gint i=0; i<count_sockets; ++i) {
		for(gint j=0; j<count_cores / count_sockets; ++j) {
			core_switch_to_mode(sockets[i][j], MODE_MAX);
		}
	}

	sleep(1);

	// MODE_MAX

	before = getTimestamp();
	sleep(5);
	after = getTimestamp();

	crows = db_getNodeMetric("power", host, before, after, &p_nodeRow);

	gdouble power_max = mi_getMeanPowerConsumption(&p_nodeRow, crows);

	free(p_nodeRow);

	// now reduce the core mode stepwise per socket
	gdouble power[count_sockets][count_cores / count_sockets];

	for(gint i=0; i<count_sockets; ++i) {
		for(gint j=0; j<count_cores / count_sockets; ++j) {
			core_switch_to_mode(sockets[i][j], MODE_MIN);
			sleep(1);

			before = getTimestamp();
			sleep(5);
			after = getTimestamp();

			crows = db_getNodeMetric("power", host, before, after, &p_nodeRow);

			power[i][j] = mi_getMeanPowerConsumption(&p_nodeRow, crows);

			free(p_nodeRow);
		}
	}

	// free cores
	for(gint i=0; i<count_sockets; ++i) {
		for(gint j=0; j<count_cores / count_sockets; ++j) {
			core_unref(sockets[i][j]);
		}
	}

	g_print("Ondemand power: %f Watt\n", power_ondemand);
	g_print("MODE_MAX power: %f Watt\n", power_max);
	g_print("Stepwise mode increase power: \n");

	for(gint i=0; i<count_sockets; ++i) {
		g_print("Socket %d: | ", i);
		for(gint j=0; j<count_cores / count_sockets; ++j) {
			g_print("%f |", power[i][j]);
		}
		g_print("\n");
	}

	gdouble socket_power_difference = (power_max - power_ondemand) / count_sockets;
	gdouble core_power_difference = socket_power_difference / (count_cores / count_sockets);

	/*
	 * Ondemand power: 106.006189 Watt
	 * MODE_MAX power: 112.700650 Watt
	 *
	 * Socket 0: | 112.809486 |112.911827 |112.878660 |113.027087 |112.888881 |112.981664 |112.923996 |112.942050 |113.009683 |113.072170 |113.067982 |109.172138 |
	 * Socket 1: | 109.033349 |109.154803 |109.065982 |109.201076 |109.115118 |109.107627 |109.015099 |109.078030 |108.986398 |109.156185 |109.109622 |105.165113 |
	 *
	 * socket_power_difference = 6.694461 Watt
	 * core_power_difference = 0.06973396875 Watt
	 *
	 */

	// assume power[i][j] > power[i][j+1], since more cores are in a lower frequency
	// check if there is linear decrease in power consumption, not only a decrease when all cores are at minimum frequency on a socket

	for(gint i=0; i<count_sockets; ++i) {
		gdouble last = power[i][0];
		for(gint j=1; j<count_cores / count_sockets; ++j) {

			// check if the decrease is in the expected rank +- 10 %
			if ((last - core_power_difference) * 1.1 > power[i][j] || (last - core_power_difference) * 0.9 < power[i][j]) {
				g_print("Reducing frequency of core %d on socket %d does result in the expected power consumption of [%f, %f]. Assuming shared voltage regulator.\n", j, i, (last - core_power_difference) * 0.9, (last - core_power_difference) * 1.1);
				return FALSE;
			}

			last = power[i][j];
		}
	}



#else
	g_print(
			"Frequency per core not distincable without power measurement! Assuming possible...\n");
	return TRUE;

#endif

}

int main(gint argc, gchar** argv) {

	if (argc == 1) {
		g_print("Please provide a filename for config file to be written!\n");
		return 1;
	}

	/* default settings */
	glong interval = 200; // 200 microseconds

	/* create lists of devices for this host */
	GList * disks = get_disk_device_names();

	GList * nics = get_nic_device_names();

	gint count_sockets;
	GHashTable * socketMapping = get_core_to_socket_mapping(&count_sockets);

	if (socketMapping == NULL) {
		/* free temporarily allocated lists */
		for (GList* it = g_list_first(disks); it != NULL; it = it->next)
			g_free((gchar*) it->data);
		g_list_free(disks);

		for (GList* it = g_list_first(nics); it != NULL; it = it->next)
			g_free((gchar*) it->data);
		g_list_free(nics);

		g_print("Exit now.\n");

		return 1;
	}

	/* distinct startup times for devices (assume same startup times for each kind of device) */

	glong * startup_times = get_startup_times(disks, nics);

	/* distict if frequency is adjustable per core */

	gboolean frequency_per_core = get_frequency_per_core(count_sockets,
			socketMapping);

	/* create config for this host */
	Config *config = config_new(count_sockets, socketMapping, disks, nics,
			startup_times, frequency_per_core, interval);

	/* write config to file provided as commandline arg */
	config_write(config, argv[1]);

	config_unref(config);

	/* free temporarily allocated lists */
	for (GList* it = g_list_first(disks); it != NULL; it = it->next)
		g_free((gchar*) it->data);
	g_list_free(disks);

	for (GList* it = g_list_first(nics); it != NULL; it = it->next)
		g_free((gchar*) it->data);
	g_list_free(nics);

	/* array with startup times is freed in config_unref */

	return 0;

}
