/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * ConstrainedResource.c
 *
 *  Created on: Jul 11, 2011
 *      Author: Timo Minartz
 */

#include "ConstrainedResource.h"
#include "print.h"

ConstrainedResource* constrained_resource_new(glong interval) {
	ConstrainedResource* cr = g_malloc(sizeof(ConstrainedResource));

	cr->resources = NULL;
	cr->interval = interval;
	cr->finalize = FALSE;

	return cr;
}

void constrained_resource_unref(ConstrainedResource * cr) {

	if (cr != NULL) {
		g_list_free(cr->resources);
		g_free(cr);
	}
}

void constrained_resource_register(ConstrainedResource * cr,
		Resource * resource) {

	if (cr != NULL && resource != NULL)
		cr->resources = g_list_prepend(cr->resources, resource);
}

void constrained_resource_unregister(ConstrainedResource * cr,
		Resource * resource) {
	if (cr != NULL && resource != NULL)
		cr->resources = g_list_remove(cr->resources, resource);
}

Mode constrained_resource_process_list(ConstrainedResource *cr,
		glong time_passed) {

	Mode mode = MODE_MAX;

	if (cr == NULL)
		return mode;

	// empty list
	if (g_list_first(cr->resources) == NULL)
		return mode;

	Mode count_per_mode[COUNT_MODES] = { 0 }; // Occurrence count of each mode

	// iterate over all resources to find the lowest possible mode
	for (GList* it = g_list_first(cr->resources); it != NULL; it = it->next) {

		Resource * resource = (Resource*) it->data;

		glong startup_time = resource_get_startup_time(resource);

		count_per_mode[resource_process_list(resource, startup_time,
				time_passed)]++;
	}

	for (mode = 1; mode < COUNT_MODES; ++mode) {
		if (count_per_mode[mode] > 0)
			break;
	}

	// now mode contains the lowest possible mode from all resources or MODE_MAX if no resources in the list

	return mode;
}
/**
 * Switches the ConstrainedResource to the desired mode. If already in this mode, nothing happens. Otherwise the switch is initiated for all Resources calling resource_switch_to_mode().
 * @param cr ConstrainedResource to operate on
 * @param mode Mode to switch to
 */
void switch_to_mode(ConstrainedResource * cr, Mode mode) {

	if (cr == NULL)
		return;

	for (GList* it = g_list_first(cr->resources); it != NULL; it = it->next) {
		Resource * resource = (Resource*) it->data;

		resource_switch_to_mode(resource, mode);
	}
}

void constrained_resource_process(ConstrainedResource * cr) {

	if (cr == NULL)
		return;

	registerThread(g_thread_self(), "processing thread");

	// in microseconds
	glong interval = cr->interval;

	Mode last_mode, mode;

	last_mode = MODE_UNKNOWN;

	// to hit the interval, we wait interval - duration for the last update
	glong waitTime = 0; // in microseconds
	GTimer *timer = g_timer_new();

	while (!cr->finalize) {

		g_usleep((gulong) waitTime);

		g_timer_start(timer);

		mode = constrained_resource_process_list(cr, interval);

		if (mode != last_mode) {
			last_mode = mode;

			if (mode == MODE_UNKNOWN)
				break;

			switch_to_mode(cr, mode);
		}

		// waitTime is the difference between the interval time and the elapsed time. If waitTime is negative, the interval is too small

		// g_timer_elapsed returns elapsed time in seconds
		waitTime = interval - (glong) (g_timer_elapsed(timer, NULL) * 1000000);
		debug(3, "processing wait time = %ld\n", waitTime);
		if (waitTime < 0) {
			waitTime = 0;
			info("Adjusted negative wait time: If occurs frequently, the processing interval may be too small.\n");
		}
	}

	g_timer_destroy(timer);
}
