/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Core.c
 *
 *  Created on: Apr 5, 2011
 *  Author: Timo Minartz
 */

#include "Core.h"

#include "print.h"
#include "Config.h"

#include <stdlib.h> /* for atoi*/
#include <string.h> /* for strcmp */
#include <cpufreq.h> /* for managing p-states of the core */
#include <stdio.h> /* for snprintf */
#include <glib.h> /* for hashtable */

// TODO: switch to these structures
//struct corefreq_available_governors * available_governors = NULL;
//struct corefreq_available_frequencies * available_frequencies = NULL;

gint set_available_frequencies(Core *core) {

	gchar * buffer;
	gsize length;
	GError *error = NULL;
	gchar ** str_frequencies;

	gchar
			frequency_file[strlen(
					"/sys/devices/system/cpu/cpu99/cpufreq/scaling_available_frequencies")
					+ 1];
	g_snprintf(
			frequency_file,
			strlen(
					"/sys/devices/system/cpu/cpu99/cpufreq/scaling_available_frequencies")
					+ 1,
			"/sys/devices/system/cpu/cpu%d/cpufreq/scaling_available_frequencies",
			core->id);

	if (g_file_get_contents(frequency_file, &buffer, &length, &error)) {
		str_frequencies = g_strsplit(buffer, " ", 20);
	} else {
		debug(1, "Error reading frequencies from file: %s\n", error->message);
		g_error_free(error);
		return ERROR_CORE_READ_FREQUENCIES;
	}

	g_free(buffer);

	core->count_frequencies = g_strv_length(str_frequencies) - 1;

	Frequency * frequencies = malloc(
			core->count_frequencies * sizeof(Frequency));
	for (gint i = 0; i < core->count_frequencies; ++i)
		frequencies[i] = atoi(str_frequencies[i]);

	core->frequencies = frequencies;

	if ((str_frequencies[0])[3] == '1') {
		// indicates turboboost
		core->has_turboboost = TRUE;
	} else {
		core->has_turboboost = FALSE;
	}

	g_strfreev(str_frequencies);

	return 0;

}

gint change_governor(Core *core, Governor governor) {

	if (core == NULL)
		return CORE_NOT_INITIALIZED;

	if (strcmp(core->governor, governor) != 0) {

		gint ret = cpufreq_modify_policy_governor(core->id, governor);
		if (ret != 0) {
			debug(1, "Cannot set governor %s for core %d\n", governor, core->id);
			return ret;
		}

		core->governor = governor;
	}

	return 0;
}

gint change_frequency(Core *core, Frequency frequency) {

	if (core == NULL)
		return CORE_NOT_INITIALIZED;

	if (frequency != core->frequency) {
		gint ret = change_governor(core, GOV_USERSPACE);

		if (ret != 0)
			return ret;

		ret = cpufreq_set_frequency(core->id, frequency);
		if (ret != 0) {
			debug(1, "Cannot set frequency %d for core %d.\n", frequency, core->id);
			return ret;
		}

		core->frequency = frequency;
	}

	return 0;
}

Mode core_get_mode(Core *core) {

	if (core == NULL)
		return MODE_UNKNOWN;

	return core->mode;
}

/**
 * Sets the current Mode of this Core
 * @param core Core to set mode of
 * @param mode Mode to set Core to
 */
void core_set_mode(Core *core, Mode mode) {

	if (core == NULL)
		return;

	core->mode = mode;

	debug(1, "Core %d switched to mode %s\n", core->id, MODE_TO_STRING(mode));
}

/**
 * Switches the Core to the desired mode. If already in this mode, nothing happens. Otherwise the switch is initiated using cpufreq.
 * @param core Core to operate on
 * @param mode Mode to switch to
 */
void core_switch_to_mode(Core * core, Mode mode) {

	if (core == NULL)
		return;

	if (core_get_mode(core) != mode) {

		gint i = 0;

		if (mode == MODE_TURBO) {
			i = 0;
		} else if (mode == MODE_MAX) {
			if (core_has_turboboost(core))
				i = 1;
			else
				i = 0;
		} else if (mode == MODE_MED) {
			i = core->count_frequencies / 2;
		} else if (mode == MODE_MIN || mode == MODE_UNUSED) {
			i = core->count_frequencies - 1;
		}

		gint ret = change_frequency(core, core->frequencies[i]);

		if (ret == 0) {
			core_set_mode(core, mode);
		} else {
			error("Executing core change to mode %s failed for core %d.\n", MODE_TO_STRING(mode), core->id);
		}
	}
}

/**
 * Switches this instance of Core to the ondemand Governor.
 * After method call the internal Mode is set to MODE_DEFAULT.
 * @param core Core to switch frequency of
 * @return 0 if succeeded and CORE_NOT_INITIALIZED otherwise
 */
gint core_standard(Core *core) {

	if (core == NULL)
		return CORE_NOT_INITIALIZED;

	gint ret = 0;

	if (core_get_mode(core) != MODE_DEFAULT) {
		// change to ondemand governor for this cores

		ret = change_governor(core, GOV_ONDEMAND);
		if (ret != 0)
			return ret;

		core_set_mode(core, MODE_DEFAULT);
	}

	return ret;
}

State core_get_state(Core *core) {
	return cpufreq_get_freq_kernel(core->id);
}

gboolean core_has_turboboost(Core *core) {

	if (core == NULL)
		return FALSE;

	return core->has_turboboost;
}

Core* core_new(gint id) {

	Core *core = g_malloc(sizeof(Core));

	if (core == NULL) {
		debug(1, "Cannot allocate memory for core.\n");
		return NULL;
	}

	core->id = id;

	gint ret = set_available_frequencies(core);

	if (ret != 0) {
		g_free(core);
		return NULL;
	}

	core->mode = MODE_DEFAULT;
	core->governor = GOV_ONDEMAND; // assert ondemand governor for core
	core->frequency = FREQ_UNKNOWN;

	ret = core_standard(core);

	if (ret != 0) {
		g_free(core->frequencies);
		g_free(core);
		return NULL;
	}

	core->startup_time = 200;

	debug(1, "Allocated new core with id %d\n", id);

	return core;
}

void core_unref(Core *core) {

	if (core != NULL) {

		core_standard(core);

		debug(1, "Unref core with id %d\n", core->id);

		g_free(core->frequencies);
		g_free(core);

		core = NULL;
	}
}
