/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Core.h
 *
 *  Created on: Apr 5, 2011
 *  Author: Timo Minartz
 */

#ifndef CORE_H_
#define CORE_H_

typedef struct Core Core;

#include <glib.h>
#include "defines.h"

/**
 * Encapsulates the interface to the hardware cores.
 */
struct Core {
	/**
	 * Identifier of the Core
	 */
	gint id;
	/**
	 * Count available frequencies;
	 */
	gint count_frequencies;
	/**
	 * Available frequencies for this Core
	 */
	Frequency* frequencies;
	/**
	 * Current Frequency of this Core
	 */
	Frequency frequency;
	/**
	 * Current Governor of this Core
	 */
	Governor governor;
	/**
	 * Current Mode of this Core
	 */
	Mode mode;
	/**
	 * Defines if this Core has turboboost feature
	 */
	gboolean has_turboboost;
	/**
	 * Time in microseconds needed for a Core to be in MODE_MAX
	 */
	glong startup_time;
};

/**
 * Returns the current State (== Frequency) of this Core in kHz or 0 in case of failure.
 * \param core Core to get frequency from
 * \return State of Core or 0 if failure
 */
State core_get_state(Core *core);

/**
 * Returns the current Mode of this Core
 * @param core Core to get mode of
 * @return Mode of Core
 */
Mode core_get_mode(Core *core);

/**
 * Returns if this Core has an internal TurboBoost like feature
 * @param core Core to get turboboost feature from
 * @return if the Core has an internal TurboBoost like feature
 */
gboolean core_has_turboboost(Core *core);

/**
 * Switches the Core to the desired mode. If already in this mode, nothing happens. Otherwise the switch is initiated using cpufreq.
 * @param core Core to operate on
 * @param mode Mode to switch to
 */
void core_switch_to_mode(Core * core, Mode mode);

/**
 * Constructor.
 * Creates a new instance of Core with the specified identifier.
 * Reads the available frequencies from /sys/devices/system/core/core$id$/corefreq/scaling_available_frequencies.
 * The instance of the constructed Core object is registered.
 * The internal thread for the ResourceManager is created.
 * @param id Identifier for this Core
 * @return new Instance of Core with the speciefied identifier in MODE_DEFAULT
 */
Core* core_new(gint id);

/**
 * Destructor.
 * Joins the internal thread for the ResourceManager.
 * Frees the allocated memory.
 */
void core_unref(Core *core);

#endif /* CORE_H_ */
