/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * server_eed.c
 *
 *  Created on: Dec 6, 2010
 *      Author: Timo Minartz
 */

#include "Daemon.h"
#include <string.h>
#include "print.h"
#include "Config.h"
#include "ResourcesTable.h"

/**
 * Nested ifdefs explanation:
 * - enable tracing in general
 * - define tracing mode (HDTrace / DBConnector)
 * - enable likwid memory bandwidth tracing
 */
#ifdef ENABLE_TRACE
#include <unistd.h> /* for gethostname */
#include <stdlib.h> /* for getenv and atol */

/**
 * hostname needed for tracing
 */
static gchar hostname [8];

static gboolean finalizeTracingThread = FALSE;

/**
 * separate tracing thread
 */
static GThread * hdTraceThread;

/**
 * tracing started?
 */
static gboolean tracing = FALSE;

/**
 * If compiled against HDTrace
 */
#ifdef HAVE_HDTRACE
#include "hdTopo.h"
#include "hdStats.h"

static hdStatsGroup *group = NULL;
static hdTopology *myTopology = NULL;
static hdTopoNode *myTopoNode = NULL;
#elif HAVE_DB
// a dummy. Otherwise, some more if blocks needed
#define hdStatsGroup gchar
#endif

/**
 * If have an modified likwid library, we can also trace the memory bandwidth
 */
#ifdef HAVE_LIKWID
#include "eeLikwid.h"
#include "Core.h"

static gint count_sockets = 0;
#endif

/**
 * If compiled against DBConnector
 */
#ifdef HAVE_DBC
#include "DBConnector.h"
#include "MetricInterpreter.h"
uint64_t tracing_start = 0;

#endif

#endif

/**
 * global ResourcesTable, containing all Resources
 */
static ResourcesTable * rt = NULL;

/**
 * global Config
 */
static Config * config = NULL;

/**
 * a global mutex, needed for thread synchronisation (e.g. initialization in ee_init should be done only once)
 */
static GStaticMutex ee_daemon_mutex = G_STATIC_MUTEX_INIT;

/**
 * Path to config file
 */
extern gchar * config_file_path;

void ee_init(gint tag, gint rank, const gchar * prgname, gint core_id) {

	g_static_mutex_lock (&ee_daemon_mutex);

	info(
			"Rank %d with tag %d initializes Daemon for program %s on core %d.\n", rank, tag, prgname, core_id);

	if (rt == NULL) {

		debug(1, "Rank %d initializes Resources Table.\n", rank);

		if (config_file_path == NULL) {
			error("Please specify config file first!\n");
			g_static_mutex_unlock(&ee_daemon_mutex);

			return;
		}

		config_read(&config, config_file_path);

		if (config == NULL) {
			error(
					"Error reading config file: %s\nAbort Initialization.\n", config_file_path);
			g_static_mutex_unlock(&ee_daemon_mutex);

			return;
		}

		rt = config_create_resources_table(config);

		if (config->interval == 0) {
			info("Processing disabled.\n");

			/*
			 * Resources will be registered and request will be generated - but not processed.
			 */

		} else {
			info("Processing interval = %ld microseconds.\n", config->interval);

			// start threads for processing of resources requests

			resources_table_start_process_resources(rt);

		}

		// register resources for this combination of tag and rank
		resources_table_register(rt, tag, rank, core_id);

#ifdef ENABLE_TRACE

		gchar * traceIntervalString = getenv("EED_TRACE_INTERVAL");

		if (traceIntervalString == NULL) {
			// use default tracing interval 100 milliseconds
			traceInterval = 100000;
		} else {
			traceInterval = atol(traceIntervalString);
		}

		if (traceInterval == 0) {
			info("[trace] Tracing disabled.\n");
		} else if (traceInterval < 1000) {
			info("[trace] Trace interval %ld microseconds is too small. Use at least 1000 microseconds.\n", traceInterval);
			info("[trace] Tracing disabled.\n");
		} else {

			info("[trace] Trace interval = %ld microseconds.\n", traceInterval);

			// only start tracing if at least one device is initiated
			if(rt != NULL && rt->registrations > 0) {

				init_tracing(prgname);

				hdTraceThread = g_thread_create((GThreadFunc) & do_tracing, NULL, TRUE, NULL);

				if (hdTraceThread == NULL) {
					error("Tracing thread could not be created.\n");
				}

				registerThread(hdTraceThread, "trace");

				tracing = TRUE;
			}
		}
#endif

	} else {
		// register all resources for this combination of tag and rank
		resources_table_register(rt, tag, rank, core_id);
	}

	g_static_mutex_unlock (&ee_daemon_mutex);

}

void ee_dev_mode(int device_id, int mode_id, int tag, int rank) {

	if (rt == NULL) {
		error("Call ee_init() first!\n");
		return;
	}

	if (IS_DEV_ALL(device_id)) {

		// Specified device_id is one out of CORE_ALL_ID, NIC_ALL_ID or DISK_ALL_ID

		GList * id_resources = resources_table_get_resources_by_id(rt,
				device_id);

		for (GList* it = g_list_first(id_resources); it != NULL; it = it->next) {
			Resource * resource = (Resource *) it->data;

			// register this combination of tag and rank to this resource if not already registered
			resource_register(resource, tag, rank);

			resource_mode_request(resource, tag, rank, mode_id);
		}

	} else if (IS_MY_ID(device_id)) {

		// Specified device_id is one out of MY_CORE_IDS, MY_NIC_IDS or id == MY_DISK_IDS

		GList * id_resources = resources_table_get_resources_by_user_and_type(
				rt, tag, rank, device_id);

		for (GList* it = g_list_first(id_resources); it != NULL; it = it->next) {
			Resource * resource = (Resource *) it->data;
			resource_mode_request(resource, tag, rank, mode_id);
		}

	} else {

		// Specified device_id is one specific CORE_ID, NIC_ID or DISK_ID

		Resource * resource = resources_table_get_resource_by_id(rt, device_id);

		if(resource != NULL) {
			resource_mode_request(resource, tag, rank, mode_id);
		} else {
			error("Unknown device id: %d for ee_dev_mode!\n", device_id);
		}
	}

}

void ee_dev_mode_in(int device_id, int mode_id, int secs, int tag, int rank) {

	if (rt == NULL) {
		error("Call ee_init() first!\n");
		return;
	}

	glong micros = secs * 1000000;

	if (IS_DEV_ALL(device_id)) {

		GList * id_resources = resources_table_get_resources_by_id(rt,
				device_id);

		for (GList* it = g_list_first(id_resources); it != NULL; it = it->next) {
			Resource * resource = (Resource *) it->data;
			resource_time_request(resource, mode_id, tag, rank, micros);
		}
	} else if (IS_MY_ID(device_id)) {

		GList * id_resources = resources_table_get_resources_by_user_and_type(
				rt, tag, rank, device_id);

		for (GList* it = g_list_first(id_resources); it != NULL; it = it->next) {
			Resource * resource = (Resource *) it->data;
			resource_time_request(resource, mode_id, tag, rank, micros);
		}

	} else {

		Resource * resource = resources_table_get_resource_by_id(rt, device_id);

		resource_time_request(resource, mode_id, tag, rank, micros);
	}

}

void ee_finalize(gint tag, gint rank) {

	g_static_mutex_lock (&ee_daemon_mutex);

	if (rt == NULL) {
		error("Call ee_init() first!\n");
		g_static_mutex_unlock (&ee_daemon_mutex);
		return;
	}

	/* unregister and free all devices in hash table */

	resources_table_unregister(rt, tag, rank);

	//TODO: move mutex lock to here?

	if (resources_table_get_count_registrations(rt) == 0) {

		if (config->interval > 0) {

			// stop threads for processing of resources requests (blocking)

			resources_table_stop_process_resources(rt);

		}

#ifdef ENABLE_TRACE
		// only finalize tracing if at least one device is initiated
		if(rt != NULL && tracing == TRUE) {
			finalize_tracing();
		}
#endif

		resources_table_unref_full(rt);
		rt = NULL;

		config_unref(config);
		config = NULL;

	}

	g_static_mutex_unlock (&ee_daemon_mutex);

}

void ee_finalize_all(gint tag) {

	g_static_mutex_lock (&ee_daemon_mutex);

	if (rt == NULL) {
		error("Call ee_init() first!\n");
		g_static_mutex_unlock (&ee_daemon_mutex);
		return;
	}

	/* unregister and free all devices in hash table */

	resources_table_unregister_tag(rt, tag);

	if (resources_table_get_count_registrations(rt) == 0) {

		if (config->interval > 0) {

			// stop threads for processing of resources requests (blocking)

			resources_table_stop_process_resources(rt);

		}

#ifdef ENABLE_TRACE
		// only finalize tracing if at least one device is initiated
		if(rt != NULL && tracing == TRUE) {
			finalize_tracing();
		}
#endif

		resources_table_unref_full(rt);
		rt = NULL;

		config_unref(config);
		config = NULL;
	}

	g_static_mutex_unlock (&ee_daemon_mutex);

}

#ifdef ENABLE_TRACE
void init_tracing(const gchar *prgname) {

	/* set hostname for tracing */
	gethostname(hostname, 8);

#ifdef HAVE_DBC
	gint query_puffer_size = resources_table_get_count_resources(rt); // count of entries per trace iteration (8 cores + nic + disk = 10 modes)
#ifdef ENABLE_STATE_TRACING
	query_puffer_size *= 2; // double puffer, we are also tracing the states...
#endif
	if(traceInterval < 10000) // increase puffer for small trace intervals
	query_puffer_size *= 10;

	db_initConnection(query_puffer_size);
#endif

#ifdef HAVE_HDTRACE
	const gchar *levels[] = {"Host"};

	myTopology = hdT_createTopology(prgname, levels, 1);

	/* create topology node */
	const gchar *path[] = {hostname};
	myTopoNode = hdT_createTopoNode(myTopology, path, 1);

	/* delete already existing trace file */
	GString * tracefilename = g_string_new("");

	g_string_printf(tracefilename, "/tmp/%s_%s_eed.stat", prgname, hostname);

	debug(1, "[trace] Removing tracefile %s if exists.\n", tracefilename->str);

	remove(tracefilename->str);

	/* add prefix /tmp/ to statistics file and do not abort if file already exists */
	group = hdS_createGroupWithFilenamePrefixAndAbortIfExists("eed", myTopoNode, 1,"/tmp/",0);
#elif HAVE_DBC
	gchar * group = NULL;
	tracing_start = getSystemTimestamp();
#endif

	resources_table_init_tracing(rt, group);

#ifdef HAVE_LIKWID

	// create an array of core ids, one core from every socket.

	GHashTable * socketMapping = config->socketMapping;
	gint count_cores = g_hash_table_size(socketMapping);

	count_sockets = config->countSockets;

	gint trace_cores[count_sockets];

	gint j = 0;
	for(gint i=0; i<count_sockets; ++i) {
		for(; j<count_cores; ++j) {
			if(GPOINTER_TO_INT(g_hash_table_lookup(socketMapping, GINT_TO_POINTER(j))) == i) {
				// first core with id j and socket i
				trace_cores[i] = j;
				debug(1, "[trace] Tracing core %d for socket %d bandwidth\n", j, i);
				gchar name[13];
				g_snprintf(name, 12, "EED_SOCKET%d_BW", i);
#ifdef HAVE_HDTRACE
				// create entry in stats
				hdS_addValue(group, name, DOUBLE, "MByte/s", "EED_BW");
				debug(1, "hdS_addValue(group, %s, DOUBLE, MByte/s, EED_BW)\n", name);
#endif
#ifdef HAVE_DBC
				// create metric in database
				db_createNodeMetric(name, hostname);
#endif
				break;
			}
		}
	}

	eeLikwid_init(count_sockets, trace_cores);

#endif

#ifdef HAVE_HDTRACE
	hdS_commitGroup(group);

	hdS_enableGroup(group);
#endif

#ifdef HAVE_DBC
	info("[trace] Init tracing for program \"%s\" with QueryPuffer size %d.\n", prgname, query_puffer_size);
#else
	info("[trace] Init tracing for program \"%s\".\n", prgname);
#endif

}

void do_tracing() {

	debug(1, "Start tracing.\n");

#ifdef HAVE_LIKWID
	debug(1, "LIKWID enabled.\n");
	eeLikwid_start();
#endif

	// to hit the interval, we wait traceInterval - duration for the last tracing step
	glong waitTime = 0;// in microseconds
	GTimer *timer = g_timer_new();

	while (!finalizeTracingThread) {

		g_usleep((gulong) waitTime);

		g_timer_start(timer);

#ifdef HAVE_DBC
		uint64_t timestamp = getSystemTimestamp();
#ifndef HAVE_HDTRACE
		gchar * group = NULL;
#endif
#endif

		resources_table_do_tracing(rt, group);

#ifdef HAVE_LIKWID
		gdouble *bw = g_malloc(count_sockets * sizeof(gdouble));
		eeLikwid_end(&bw);

		for(gint i=0; i<count_sockets; ++i) {
#ifdef HAVE_HDTRACE
			hdS_writeDoubleValue(group, bw[i]);
#endif
#ifdef HAVE_DBC
			gchar metric[12];
			g_snprintf(metric, 11, "eed_socket%d_bw", i);
			db_pushNodeMetric(metric, hostname, timestamp, bw[i]);
#endif
			debug(1, "Socket %d: %f MByte/s\n", i, bw[i]);
		}

		g_free(bw);

		eeLikwid_start(bw);
#endif

		g_timer_stop(timer);

		// waitTime is the difference between the interval time and the elapsed time. If waitTime is negative, we have a problem here.
		// g_timer_elapsed returns elapsed time in seconds
		waitTime = traceInterval - (glong) ( g_timer_elapsed(timer, NULL) * 1000000 );
		debug(2, "waitTime = %ld\n", waitTime);
		if(waitTime < 0) {
			waitTime = 0;
			info("Adjusted negative wait time.\n");
		}
	}

	g_timer_destroy(timer);

	finalizeTracingThread = FALSE;

	debug(1, "Finished tracing.\n");
}

void finalize_tracing() {

	finalizeTracingThread = TRUE;

	// wait for tracing thread to finish
	g_thread_join(hdTraceThread);

	debug(1, "Joined tracing thread.\n");

#ifdef HAVE_HDTRACE
	hdS_disableGroup(group);

	hdS_finalize(group);
#endif

#ifdef HAVE_LIKWID
	eeLikwid_finalize();
#endif

#ifdef HAVE_HDTRACE
	/* destroy UtilTrace object */
	hdT_destroyTopoNode(myTopoNode);
	hdT_destroyTopology(myTopology);
#endif

#ifdef HAVE_DBC
	db_finishConnection();
	uint64_t tracing_stop = getSystemTimestamp();
	info("[trace] From %lld to %lld = %f seconds.\n", tracing_start, tracing_stop, mi_getTimestampInSec(tracing_stop-tracing_start));
#endif

	info("[trace] Finalized tracing.\n");

	tracing = FALSE;

}
#endif
