/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * server_eed.h
 *
 *  Created on: Jan 10, 2011
 *  Author: Timo Minartz
 */

#ifndef SERVER_EED_H_
#define SERVER_EED_H_

#include <glib.h>

/**
 * Implementation of the initialization process on the server side.
 * Based on the hostname, a new Config is created and the corresponding DeviceTable is initialized.
 * All devices in the DeviceTable are registered at the corresponding ResourceManager.
 * If compiled with tracing support, the tracing is also initialized.
 * @param tag Tag to identify the caller
 * @param rank MPI rank to identify the caller
 * @param prgname Name of the MPI program (needed for tracing), otherwise ignored
 * @param core_id ID of the processor core the process is pinned to
 */
void ee_init(gint tag, gint rank, const gchar * prgname, gint core_id);

/**
 * Initiates a mode change of the specified device.
 * Based on the device_id, the corresponding device is fetched from the DeviceTable.
 * A new request is created for the corresponding ResourceManager.
 * If the Mode change will happen depends on the requests in the queue and on the users of this device.
 * @param device_id ID of the device to change Mode of
 * @param mode_id Mode to change to
 * @param tag Tag to identify the caller
 * @param rank MPI rank to identify the caller
 */
void ee_dev_mode(gint device_id, gint mode_id, gint tag, gint rank);

/**
 * Initiates a mode change of the specified device if possible in the specified idle time.
 * Based on the device_id, the corresponding device is fetched from the DeviceTable.
 * A new request is created for the corresponding ResourceManager.
 * If the Mode change will happen depends on the requests in the queue and on the users of this device.
 * @param device_id ID of the device to change Mode of
 * @param secs Seconds this device is assumed to be idle
 * @param tag Tag to identify the caller
 * @param rank MPI rank to identify the caller
 */
void ee_dev_idle_for(gint device_id, gint secs, gint tag, gint rank);

/**
 * Initiates a mode change of the specified device if possible in the specified idle time.
 * This method is only a wrapper around ee_dev_idle_for.
 * @param device_id ID of the device to change Mode of
 * @param secs Seconds this device should be ready again
 * @param tag Tag to identify the caller
 * @param rank MPI rank to identify the caller
 */
void ee_dev_ready_in(gint device_id, gint secs, gint tag, gint rank);

/**
 * Unregisters and frees all devices in the DeviceTable.
 * If tracing is enabled, it is also finalized.
 * @param tag Tag to identify the caller
 * @param rank MPI rank to identify the caller
 */
void ee_finalize(gint tag, gint rank);

/**
 * Unregisters and frees all devices in the DeviceTable marked with the specified tag.
 * @param tag Identification for devices to finalize
 */
void ee_finalize_all(gint tag);

#ifdef ENABLE_TRACE

/**
 * Trace interval in microseconds
 * 1000000 microseconds = 1 second
 */
static glong traceInterval;

/**
 * Initializes statistics group for hdTrace.
 * @param prgname Name of the program to trace
 */
void init_tracing(const gchar* prgname);

/**
 * Start point for tracing thread.
 * While tracing, echo every traceInterval microseconds the mode of each device to statistics group
 */
void do_tracing();

/**
 * Finalizes the statistics group.
 */
void finalize_tracing();
#endif

#endif /* SERVER_EED_H_ */
