/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Disk.c
 *
 *  Created on: Feb 22, 2011
 *  Author: Timo Minartz
 */

#include "Disk.h"

#include "print.h"
#include "internal_disk.h"
#include <stdlib.h> /* for system */

/**
 * for executing commands
 */
void execute_command(gchar * command) {
	gint i = system(command);
	debug(1, "Executed command \"%s\" with return value %d\n", command, i);
}

/**
 * Generates remount commands for the partitions of this disk
 */
void generate_commands(Disk * disk) {

	gchar * command[] = { "/bin/mount", NULL };
	gchar * standard_out = NULL;
	gchar * standard_err = NULL;
	gint exit_status = 0;
	GError* error = NULL;
	disk->remount_commands = NULL;
	disk->remount_ee_commands = NULL;

	gboolean ret = g_spawn_sync(NULL, command, NULL, G_SPAWN_SEARCH_PATH, NULL,
			NULL, &standard_out, &standard_err, &exit_status, &error);

	if (!ret) {
		error("Unable to spawn process for getting mounts: %s\n", error->message);
		g_error_free(error);
		return;
	}

	if (standard_out != NULL) {

		gchar ** lines = g_strsplit(standard_out, "\n", 100);

		gint count = g_strv_length(lines);

		for (gint i = 0; i < count; ++i) {

			gchar ** params = g_strsplit(lines[i], " ", 10); // /dev/sda1 on / type ext3 (rw,acl,user_xattr)

			if (params[0] == NULL || !g_str_has_prefix(params[0], disk->device)) {
				g_strfreev(params);
				continue;
			}

			/*
			 * construct the command string like:
			 * mount -o remount,rw,acl,user_xattr,noatime,commit=43200 /dev/sda1 /
			 * mount -o remount,rw,acl,user_xattr /dev/sda1 /
			 */

			// remove parentheses
			gchar * opts = g_strdup(params[5]); // (rw,acl,user_xattr)
			opts[strlen(opts) - 1] = '\0';
			opts += 1;

			disk->remount_ee_commands
					= g_list_prepend(
							disk->remount_ee_commands,
							g_strconcat("mount -o remount,", opts,
									",noatime,commit=43200 ", params[0], " ",
									params[2], NULL));
			disk->remount_commands = g_list_prepend(
					disk->remount_commands,
					g_strconcat("mount -o remount,", opts, " ", params[0], " ",
							params[2], NULL));

			opts -= 1;
			opts[strlen(opts) - 1] = ')';
			g_free(opts);
			g_strfreev(params);

		}

		g_strfreev(lines);

	} else {
		error("Executing command failed: %s\n", standard_err);
	}

	g_free(standard_err);
	g_free(standard_out);

}

/**
 * Sets the Mode of the Disk
 * @param disk Disk to set Mode of
 * @param mode Mode to switch to
 */
void disk_set_mode(Disk *disk, Mode mode) {

	if (disk == NULL)
		return;

	disk->mode = mode;
}

void disk_init(Disk * disk) {

	if (disk == NULL)
		return;

	if (disk_get_mode(disk) == MODE_DEFAULT) {
		/* reduce background disk writings */

		// disable pdflush daemon
		execute_command("echo 0 > /proc/sys/vm/dirty_writeback_centisecs"); /*The pdflush writeback daemons will periodically wake up and write "old" data out to disk. This tunable expresses the interval between those wakeups, in 100'ths of a second. Setting this to zero disables periodic writeback altogether.*/
		execute_command("echo 99 > /proc/sys/vm/dirty_background_ratio"); /* Contains, as a percentage of total system memory, the number of pages at which the pdflush background writeback daemon will start writing out dirty data. */
		execute_command("echo 4320000 > /proc/sys/vm/dirty_expire_centisecs"); /* This tunable is used to define when dirty data is old enough to be eligible for writeout by the pdflush daemons. It is expressed in 100'ths of a second. Data which has been dirty in memory for longer than this interval will be written out next time a pdflush daemon wakes up.*/
		execute_command("echo 99 > /proc/sys/vm/dirty_ratio"); /* Contains, as a percentage of total system memory, the number of pages at which a process which is generating disk writes will itself start writing out dirty data. */

		// remount with writing journal every 12 hours
		for (GList* it = g_list_first(disk->remount_ee_commands); it != NULL; it
				= it->next)
			execute_command(it->data);
	}
}

State disk_get_state(Disk *disk) {
	if (disk == NULL)
		return STATE_UNKNOWN;

	return id_check_drive_power_mode_status();
}

/**
 * Switches the Disk to the desired mode. If already in this mode, nothing happens. Otherwise the switch is initiated using hdparm.
 * @param disk Disk to operate on
 * @param mode Mode to switch to
 */
void disk_switch_to_mode(Disk * disk, Mode mode) {

	if (disk == NULL)
		return;

	/*
	 * will set mount and flush options if first call
	 */
	disk_init(disk);

	Mode current_mode = disk_get_mode(disk);

	if (current_mode != mode) {

		if (mode == MODE_MAX || mode == MODE_TURBO) {

			/* check if trying to switch from MODE_MAX to MODE_TURBO which don't make sense in terms of hdparm commands */
			if (current_mode != MODE_MAX && current_mode != MODE_TURBO)
				id_flush();
		} else if (mode == MODE_MIN || mode == MODE_MED) {

			/* check if trying to switch from MODE_MIN to MODE_MED which don't make sense in terms of hdparm commands */
			if (current_mode != MODE_MIN && current_mode != MODE_MED) {

				/*
				 * -y     Force an IDE drive to immediately enter the low  power  consump-
				 * tion standby mode, usually causing it to spin down.  The current
				 * power mode status can be checked using the -C flag.
				 */
				id_flush();
				id_put_drive_in_standby();
			}
		} else if (mode == MODE_UNUSED) {
			/** Extract from hdparm:
			 * -Y     Force an IDE drive to immediately enter the  lowest  power  con-
			 *        sumption sleep mode, causing it to shut down completely.  A hard
			 *        or soft reset is required before the drive can be accessed again
			 *        (the  Linux IDE driver will automatically handle issuing a reset
			 *        if/when needed).  The current power mode status can  be  checked
			 *        using the -C flag.
			 */

			/** Alternatives to hdparm -Y **/

			/*
			 * -S Put  the  drive  into idle (low-power) mode, and also set the standby (spindown) timeout for the drive.  This timeout value is used by the drive to determine how long to wait (with no disk
			 *    activity) before turning off the spindle motor to save power.  Under such circumstances, the drive may take as long as 30 seconds to respond to a subsequent disk access, though most drives
			 *    are  much  quicker.   The  encoding of the timeout value is somewhat peculiar.  A value of zero means "timeouts are disabled": the device will not automatically enter standby mode.  Values
			 *    from 1 to 240 specify multiples of 5 seconds, yielding timeouts from 5 seconds to 20 minutes.  Values from 241 to 251 specify from 1 to 11 units of 30 minutes, yielding  timeouts  from  30
			 *    minutes  to 5.5 hours.  A value of 252 signifies a timeout of 21 minutes. A value of 253 sets a vendor-defined timeout period between 8 and 12 hours, and the value 254 is reserved.  255 is
			 *     interpreted as 21 minutes plus 15 seconds.  Note that some older drives may have very different interpretations of these values.
			 *
			 * --idle-immediate
			 *         Issue  an  ATA  IDLE_IMMEDIATE  command, to put the drive into a
			 *         lower power state.  Usually the device remains spun-up.
			 *
			 *  --idle-unload
			 *         Issue an ATA IDLE_IMMEDIATE_WITH_UNLOAD command,  to  unload  or
			 *         park the heads and put the drive into a lower power state.  Usu-
			 *         ally the device remains spun-up.
			 */

			/* sleep mode */
			id_flush();
			id_put_drive_to_sleep();
		}

		disk_set_mode(disk, mode);
	}
}

void disk_standard(Disk * disk) {

	if (disk == NULL)
		return;

	if (disk_get_mode(disk) != MODE_DEFAULT) {

		disk_switch_to_mode(disk, MODE_MAX);

		// echo standard values to vm
		execute_command("echo 500 > /proc/sys/vm/dirty_writeback_centisecs");
		execute_command("echo 10 > /proc/sys/vm/dirty_background_ratio");
		execute_command("echo 3000 > /proc/sys/vm/dirty_expire_centisecs");
		execute_command("echo 40 > /proc/sys/vm/dirty_ratio");

		// remount with standard options
		for (GList* it = g_list_first(disk->remount_commands); it != NULL; it
				= it->next)
			execute_command(it->data);

		disk_set_mode(disk, MODE_DEFAULT);
	}
}

Mode disk_get_mode(Disk *disk) {
	if (disk == NULL)
		return MODE_UNKNOWN;

	return disk->mode;
}

Disk* disk_new() {

	return disk_new_dev("/dev/sda", 200);

}

Disk * disk_new_dev(gchar * device, glong startup_time) {

	Disk *disk = g_malloc(sizeof(Disk));

	disk->device = g_strdup(device);
	disk->mode = MODE_DEFAULT; // assert default values for disk
	disk->startup_time = startup_time;

	if (id_open(disk->device) != 0) {
		g_free(disk->device);
		g_free(disk);
		debug(1, "Failed disk allocation: Cannot execute id_open on device %s\n", device);
		return NULL;
	}

	generate_commands(disk);

	disk_standard(disk);

	disk->startup_time = startup_time;

	debug(1, "Allocated new disk with device %s\n", device);

	return disk;
}

void disk_unref(Disk *disk) {

	if (disk != NULL) {

		disk_standard(disk);

		id_close();

		debug(1, "Unref disk with device %s\n", disk->device);

		g_free(disk->device);

		for (GList* it = g_list_first(disk->remount_commands); it != NULL; it
				= it->next)
			g_free(it->data);

		g_list_free(disk->remount_commands);

		for (GList* it = g_list_first(disk->remount_ee_commands); it != NULL; it
				= it->next)
			g_free(it->data);

		g_list_free(disk->remount_ee_commands);

		g_free(disk);
	}
}

