/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Nic.c
 *
 *  Created on: Feb 22, 2011
 *  Author: Timo Minartz
 */

#include "Nic.h"

#include <sys/ioctl.h> /* for ioctl */
#include <linux/sockios.h> /* for SIOCETHTOOL */
#include <string.h> /* for strncopy */
#include <errno.h> /* for error numbers */

#include "print.h"

/**
 * Sets the Mode of the Nic
 * @param nic Nic to set Mode of
 * @param mode Mode to switch to
 */
void nic_set_mode(Nic *nic, Mode mode) {

	if (nic == NULL)
		return;

	nic->mode = mode;

	debug(1, "Nic %s switched to mode %s\n", nic->device, MODE_TO_STRING(mode));
}

gint nic_execute_command(Nic *nic) {

	if (nic == NULL)
		return NIC_NOT_INITIALIZED;

	gint ret = 0; //ioctl(nic->sock, SIOCETHTOOL, &(nic->ifr));

	if (ret < 0) {
		error("Error while executing ioctl for NIC: %s\n", strerror(errno));
		return ret;
	}

	return ret;
}

gint nic_init(Nic *nic) {

	if (nic == NULL)
		return NIC_NOT_INITIALIZED;

	if (nic_get_mode(nic) == MODE_DEFAULT) {
		// switch autonegation off

		nic->edata.autoneg = AUTONEG_DISABLE;
		nic->edata.cmd = ETHTOOL_SSET;
		nic->edata.speed = SPEED_1000;

		return nic_execute_command(nic);
	}

	return 0;

}

/**
 * Switches the Nic to the desired mode. If already in this mode, nothing happens. Otherwise the switch is initiated using ethtool_cmd.
 * @param nic Nic to operate on
 * @param mode Mode to switch to
 */
void nic_switch_to_mode(Nic * nic, Mode mode) {

	if (nic == NULL)
		return;

	gint ret = nic_init(nic);

	if (ret != 0) {
		error("Initializing NIC failed while executing change to mode %s.\n", MODE_TO_STRING(mode));
		return;
	}

	Mode current_mode = nic_get_mode(nic);

	if (current_mode != mode) {

		if (mode == MODE_MAX || mode == MODE_TURBO) {

			/* check if trying to switch from MODE_MAX to MODE_TURBO which don't make sense */
			if (current_mode == MODE_MAX || current_mode == MODE_TURBO) {
				// only update the mode
				nic_set_mode(nic, mode);
				return;
			}

			nic->edata.speed = SPEED_1000;
			nic->edata.duplex = DUPLEX_FULL;
			nic->edata.cmd = ETHTOOL_SSET;
		} else if (mode == MODE_MED) {
			nic->edata.speed = SPEED_100;
			nic->edata.duplex = DUPLEX_FULL;
			nic->edata.cmd = ETHTOOL_SSET;
		} else if (mode == MODE_MIN) {
			nic->edata.speed = SPEED_10;
			nic->edata.duplex = DUPLEX_FULL;
			nic->edata.cmd = ETHTOOL_SSET;
		} else if (mode == MODE_UNUSED) {
			nic->edata.speed = SPEED_10;
			nic->edata.duplex = DUPLEX_HALF;
			nic->edata.cmd = ETHTOOL_SSET;
		}

		nic_set_mode(nic, MODE_UNKNOWN);

		debug(1, "Switching nic to %d with duplex %d.\n", nic->edata.speed, nic->edata.duplex);

		ret = nic_execute_command(nic);

		if (ret == 0)
			nic_set_mode(nic, mode);
		else
		error("Executing nic change to mode %s failed.\n", MODE_TO_STRING(mode));
	}
}

gint nic_standard(Nic *nic) {

	if (nic == NULL)
		return NIC_NOT_INITIALIZED;

	gint ret = 0;

	if (nic_get_mode(nic) != MODE_DEFAULT) {

		// if already in max mode, do nothing (because of time delay when terminating daemon)...
		if (nic_get_mode(nic) == MODE_MAX) {
			nic_set_mode(nic, MODE_DEFAULT);
			return ret;
		}

		// else switch autonegation on

		nic->edata.duplex = DUPLEX_FULL;
		nic->edata.cmd = ETHTOOL_SSET;
		nic->edata.autoneg = AUTONEG_ENABLE;
		nic->edata.speed = SPEED_1000;

		ret = nic_execute_command(nic);

		if (ret == 0)
			nic_set_mode(nic, MODE_DEFAULT);
	}

	return ret;

}

Mode nic_get_mode(Nic * nic) {
	if (nic == NULL)
		return MODE_UNKNOWN;

	return nic->mode;
}

State nic_get_state(Nic *nic) {

	if (nic == NULL)
		return STATE_UNKNOWN;

	struct ifreq ifr;
	strncpy(ifr.ifr_name, nic->device, sizeof(ifr.ifr_name));

	struct ethtool_cmd edata;
	edata.cmd = ETHTOOL_GSET;

	ifr.ifr_data = (caddr_t) &edata;

	gint ret = ioctl(nic->sock, SIOCETHTOOL, &ifr);

	if (ret != 0) {
		debug(1, "Failed to execute ioctl\n");
		return STATE_UNKNOWN;
	}

	switch (edata.speed) {
	case 10:
		return 10;
	case 100:
		return 100;
	case 1000:
		return 1000;
	default:
		return STATE_SWITCHING;
	}
}

Nic* nic_new() {
	return nic_new_dev("eth0", 200);
}

Nic* nic_new_dev(gchar *device, glong startup_time) {

	// TODO: use speed array for available speeds from nic

	gint * speed = g_malloc(3 * sizeof(gint));
	speed[0] = 10;
	speed[1] = 100;
	speed[2] = 1000;

	Nic *nic = g_malloc(sizeof(Nic));

	nic->device = g_strdup(device);
	nic->mode = MODE_DEFAULT; // assert NIC in autonegation mode
	nic->speed = speed;
	nic->count_speed = 3;

	/* initialize socket stuff */
	nic->sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);
	if (nic->sock < 0) {
		error("Error while creating socket.\n");
		g_free(nic);
		g_free(speed);
		g_free(device);
		return NULL;
	}

	strncpy(nic->ifr.ifr_name, nic->device, sizeof(nic->ifr.ifr_name));

	// first read data from device to get valid values
	nic->edata.cmd = ETHTOOL_GSET;
	nic->ifr.ifr_data = (caddr_t) &(nic->edata);

	nic_execute_command(nic);

	int ret = nic_standard(nic);

	if (ret != 0) {

		error("Error initializing NIC: nic_standard failed with return code: %d\n", ret);

		shutdown(nic->sock, SHUT_RDWR);

		g_free(nic->speed);
		g_free(nic->device);
		g_free(nic);
		return NULL;
	}

	nic->startup_time = startup_time;

	debug(1, "Allocated new NIC with device %s\n", device);

	return nic;
}

void nic_unref(Nic *nic) {

	if (nic != NULL) {

		nic_standard(nic);

		shutdown(nic->sock, SHUT_RDWR);

		debug(1, "Unref NIC with device %s\n", nic->device);

		g_free(nic->speed);
		g_free(nic->device);
		g_free(nic);

		nic = NULL;
	}
}
