/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Resource.c
 *
 *  Created on: Jul 11, 2011
 *      Author: Timo Minartz
 */

#include <glib.h>

#include "Resource.h"
#include "Core.h"
#include "Nic.h"
#include "Disk.h"
#include "Config.h"
#include "print.h"
#include <string.h> /* for strlen */

#ifdef ENABLE_TRACE
#ifdef HAVE_DBC
#include "DBConnector.h"
#include <unistd.h> /* for gethostname */
#endif
#endif

typedef struct {
	gint tag;
	gint rank;
} User;

typedef struct {
	User *user;
	Mode mode;
	glong ready_in;
	Type type;
} Request;

Resource * resource_new(gint device_id, gpointer device) {

	Resource * resource = NULL;

	if (device != NULL) {

		resource = g_malloc(sizeof(Resource));

		resource->requests = NULL;
		resource->users = NULL;
		resource->countUsers = 0;
		resource->countRequests = 0;
		resource->id = device_id;
		resource->device = device;
	}

	return resource;
}

void unref_resource(Resource * resource, gboolean full) {
	if (resource != NULL) {

		// free all requests
		for (GList* it = g_list_first(resource->requests); it != NULL; it
				= it->next)
			g_free((Request*) it->data);

		g_list_free(resource->requests);

		// free all users
		for (GList* it = g_list_first(resource->users); it != NULL; it
				= it->next)
			g_free((User*) it->data);

		g_list_free(resource->users);

		if (full) {
			gint id = resource->id;

			if (IS_CORE_ID(id)) {

				core_unref((Core *) resource->device);

			} else if (IS_NIC_ID(id)) {

				nic_unref((Nic *) resource->device);

			} else if (IS_DISK_ID(id)) {

				disk_unref((Disk *) resource->device);
			}
		}

		g_free(resource);
	}
}

void resource_unref_full(Resource * resource) {
	unref_resource(resource, TRUE);
}

void resource_unref(Resource * resource) {
	unref_resource(resource, FALSE);
}

User * resource_get_user(Resource * resource, int tag, int rank) {
	if (resource == NULL)
		return NULL;

	for (GList* it = g_list_first(resource->users); it != NULL; it = it->next) {

		User * user = (User*) it->data;

		if (user->tag == tag && user->rank == rank)
			return user;

	}

	//debug(1, "User (%d,%d) not found\n", tag, rank);

	return NULL;
}

void resource_register(Resource * resource, int tag, int rank) {
	if (resource != NULL) {

		if(resource_get_user(resource, tag, rank) != NULL)
			return; // already registered

		User * user = g_malloc(sizeof(User));
		user->tag = tag;
		user->rank = rank;
		resource->users = g_list_prepend(resource->users, user);
		resource->countUsers++;

#ifdef DEBUG
		gint id = resource->id;
		gpointer device = resource->device;

		if (IS_CORE_ID(id)) {
			debug(1, "Registered User (%d,%d) for Core %d\n", user->tag, user->rank, ((Core*) device)->id);
		} else if (IS_NIC_ID(id)) {
			debug(1, "Registered User (%d,%d) for Nic %s\n", user->tag, user->rank, ((Nic*) device)->device);
		} else if (IS_DISK_ID(id)) {
			debug(1, "Registered User (%d,%d) for Disk %s\n", user->tag, user->rank, ((Disk*) device)->device);
		}
#else
		debug(1, "Registered User (%d,%d)\n", user->tag, user->rank);
#endif

	}
}

Request * resource_get_request(Resource * resource, User * user) {

	if (resource == NULL)
		return NULL;

	for (GList* it = g_list_first(resource->requests); it != NULL; it
			= it->next) {

		Request * request = (Request*) it->data;

		if (request->user == user) {
			return request;
		}
	}

	return NULL;
}

void resource_unregister(Resource * resource, int tag, int rank) {

	if (resource != NULL) {

		User * user = resource_get_user(resource, tag, rank);

		if (user != NULL) {
			resource->users = g_list_remove(resource->users, user);
			g_free(user);
			resource->countUsers--;
#ifdef DEBUG
			gint id = resource->id;
			gpointer device = resource->device;

			if (IS_CORE_ID(id)) {
				debug(1, "Unregistered User (%d,%d) for Core %d\n", tag, rank, ((Core*) device)->id);
			} else if (IS_NIC_ID(id)) {
				debug(1, "Unregistered User (%d,%d) for Nic %s\n", tag, rank, ((Nic*) device)->device);
			} else if (IS_DISK_ID(id)) {
				debug(1, "Unregistered User (%d,%d) for Disk %s\n", tag, rank, ((Disk*) device)->device);
			}
#else
			debug(1, "Unregistered User (%d,%d).\n", tag, rank);
#endif
		}
	}
}

void resource_unregister_tag(Resource * resource, gint tag) {

	if (resource != NULL) {

		gint before = resource->countUsers;

		for (GList* it = g_list_first(resource->users); it != NULL; /* manually */) {

			User * user = (User *) it->data;

			if (user->tag == tag) {

				it = it->next;
				resource->users = g_list_remove(resource->users, user);
				g_free(user);
				resource->countUsers--;
				continue;
			}

			it = it->next;
		}

		debug(1, "Unregistered %d users with tag %d.\n", before - resource->countUsers, tag);
#ifdef DEBUG
		gint id = resource->id;
		gpointer device = resource->device;

		if (IS_CORE_ID(id)) {
			debug(1, "Unregistered %d User(s) (%d,*) for Core %d\n", before - resource->countUsers, tag, ((Core*) device)->id);
		} else if (IS_NIC_ID(id)) {
			debug(1, "Unregistered %d User(s) (%d,*) for Nic %s\n", before - resource->countUsers, tag, ((Nic*) device)->device);
		} else if (IS_DISK_ID(id)) {
			debug(1, "Unregistered %d User(s) (%d,*) for Disk %s\n", before - resource->countUsers, tag, ((Disk*) device)->device);
		}
#else
		debug(1, "Unregistered %d User(s) (%d,*).\n", before - resource->countUsers, tag);
#endif

	}
}

void resource_delete_request(Resource *resource, Request * request) {
	if (resource != NULL && resource != NULL) {
		resource->requests = g_list_remove(resource->requests, request);
		g_free(request);
		resource->countRequests--;
	}
}

void resource_mode_request(Resource *resource, int tag, int rank, Mode mode) {

	if (resource == NULL)
		return;

	// find user in registered users
	User * user = resource_get_user(resource, tag, rank);

	if (user == NULL)
		return;

	// check if we already have a request from this user, if so update, otherwise create a new request
	// this results in all previous requests being canceled!
	Request *request = resource_get_request(resource, user);

	if (request == NULL) {
		debug(1, "Create new mode request for Resource %d from user (%d,%d) to mode %s\n", resource->id, user->tag,
				user->rank, MODE_TO_STRING(mode));
		request = g_malloc(sizeof(Request));
		request->user = user;
		request->mode = mode;
		request->ready_in = -1;
		request->type = TYPE_MODE_REQUEST;
		resource->requests = g_list_prepend(resource->requests, request);
		resource->countRequests++;
	} else if (request->type == TYPE_MODE_REQUEST) {
		debug(1, "Update mode request for Resource %d from user (%d,%d) to mode %s\n", resource->id, user->tag,
				user->rank, MODE_TO_STRING(mode));
		request->mode = mode;
	} else {
		debug(1,
				"Update time request for Resource %d to state request for user (%d,%d) to mode %s\n",
				resource->id, user->tag, user->rank, MODE_TO_STRING(mode));
		request->ready_in = -1;
		request->type = TYPE_MODE_REQUEST;
		request->mode = mode;
	}

}

void resource_time_request(Resource *resource, gint mode, gint tag, gint rank,
		glong macros) {

	if (resource == NULL)
		return;

	// find user in users
	User * user = resource_get_user(resource, tag, rank);

	if (user == NULL)
		return;

	// check if we already have a request from this user, if so update, otherwise create a new request
	// this results in all previous requests being canceled!
	Request *request = resource_get_request(resource, user);

	if (request == NULL) {
		debug(1, "Create new time request for user (%d,%d) to mode %s in %ld\n",
				user->tag, user->rank, MODE_TO_STRING(mode), macros);
		request = g_malloc(sizeof(Request));
		request->user = user;
		request->mode = mode;
		request->ready_in = macros;
		request->type = TYPE_TIME_REQUEST;
		resource->countRequests++;
		resource->requests = g_list_prepend(resource->requests, request);
	} else if (request->type == TYPE_TIME_REQUEST) {
		debug(1, "Update time request for user (%d,%d) to mode %s in %ld\n",
				user->tag, user->rank, MODE_TO_STRING(mode), macros);
		request->mode = mode;
		request->ready_in = macros;
	} else {
		debug(1,
				"Update state request to time request for user (%d,%d) to mode %s in %ld\n",
				user->tag, user->rank, MODE_TO_STRING(mode), macros);
		request->ready_in = macros;
		request->mode = mode;
		request->type = TYPE_TIME_REQUEST;
	}

}

Mode resource_process_list(Resource *resource, glong startup_time,
		glong time_passed) {

	if (resource == NULL)
		return MODE_MAX;

	// if no user registered for this resource -> this resource can be turned off
	// at least the rms must be registered as a user if a job is running
	if (resource->countUsers == 0)
		return MODE_UNUSED;

	// a user is registered, but no requests (yet) -> this resource should be in MODE_MAX
	// This is especially the case when only the rms is registered (meaning an uninstrumented application running)
	// If the application is instrumented, MODE_MAX is a safe standard
	if (resource->countRequests == 0)
		return MODE_MAX;

	Mode mode = MODE_UNKNOWN; // mode to return

	Mode count_per_mode[COUNT_MODES] = { 0 }; // Occurrence count of each mode

	// iterate over all requests to find out if all want the same
	for (GList* it = g_list_first(resource->requests); it != NULL; it
			= it->next) {

		Request * request = (Request*) it->data;

		// handle time request
		if (request->type == TYPE_TIME_REQUEST) {

			//update the time
			request->ready_in = request->ready_in - time_passed;

			if (request->ready_in < 0)
				request->ready_in = 0;

			if (request->ready_in <= 2 * startup_time) {

				// if device must be ready in less than 2 times the startup time, return MODE_MAX;
				debug(1, "tag %d, rank %d causes a mode change to mode %s.\n", request->user->tag,
						request->user->rank, MODE_TO_STRING(request->mode));
				mode = request->mode;

				// update request to MODE_REQUEST
				request->ready_in = -1;
				request->type = TYPE_MODE_REQUEST;

				continue; // go on to update remaining time requests
			} else {
				// the device can sleep
				count_per_mode[MODE_UNUSED]++;
			}
		} else {
			// increase the mode occurence count
			count_per_mode[request->mode]++;
		}

	}

	if (mode == MODE_MAX)
		return mode; // we must change to ready mode because of a time request, these have a higher priority

	// iterate over the occurrence count and ignore MODE_UNKNOWN

	for (mode = 1; mode < COUNT_MODES; ++mode) {
		if (count_per_mode[mode] > 0)
			break;
	}

	// now mode contains the lowest possible mode from all devices

	return mode;
}

void resource_switch_to_mode(Resource * resource, Mode mode) {

	if (resource == NULL)
		return;

	gint id = resource->id;
	gpointer device = resource->device;

	if (IS_CORE_ID(id)) {

		core_switch_to_mode((Core*) device, mode);

	} else if (IS_NIC_ID(id)) {

		nic_switch_to_mode((Nic*) device, mode);

	} else if (IS_DISK_ID(id)) {

		disk_switch_to_mode((Disk*) device, mode);

	} else {
		//unknown device id should not occur
		debug(1, "Unknown device id: %d\n", id);
	}
}

glong resource_get_startup_time(Resource * resource) {

	if (resource == NULL)
		return 0;

	gint id = resource->id;

	if (IS_CORE_ID(id)) {

		return ((Core *) resource->device)->startup_time;

	} else if (IS_NIC_ID(id)) {

		return ((Nic *) resource->device)->startup_time;

	} else if (IS_DISK_ID(id)) {

		return ((Disk *) resource->device)->startup_time;

	} else {
		//unknown device id should not occur
		debug(1, "Unknown device id: %d\n", id);
		return 0;
	}
}

gboolean resource_is_registered_by_tag_and_rank(Resource * resource, gint tag, gint rank) {
	return resource_get_user(resource, tag, rank) != NULL;
}

#ifdef ENABLE_TRACE

void resource_init_tracing(Resource * resource, hdStatsGroup *group) {

	if (resource == NULL)
	return;

#ifdef HAVE_DBC
	/* set hostname for tracing */
	gethostname(hostname, 8);
#endif

	gint id = resource->id;

	if (IS_CORE_ID(id)) {

		Core * core = (Core *) resource->device;

		gchar mode_name[20];
		gchar state_name[20];
		g_snprintf(mode_name, 20, "EED_CORE%d_MODE", core->id);
		g_snprintf(state_name, 20, "EED_CORE%d_STATE", core->id);

#ifdef HAVE_HDTRACE
		// create value in stats file
		hdS_addValue(group, mode_name, INT32, "MODE", "EED_MODE");
#ifdef ENABLE_STATE_TRACING
		hdS_addValue(group, state_name, INT32, "kHz", "EED_FREQ");
#endif
#endif
#ifdef HAVE_DBC
		// create metric in database
		db_createNodeMetric(mode_name, hostname);
#ifdef ENABLE_STATE_TRACING
		db_createNodeMetric(state_name, hostname);
#endif
#endif

		debug(1, "Initalized tracing for Core %d\n", id);

	} else if (IS_NIC_ID(id)) {

		Nic * nic = (Nic *) resource->device;

		gchar mode_name[20 + strlen(nic->device)];
		gchar state_name[20 + strlen(nic->device)];
		g_snprintf(mode_name, 19 + strlen(nic->device), "EED_NIC_%s_MODE", nic->device);
		g_snprintf(state_name, 19 + strlen(nic->device), "EED_NIC_%s_STATE", nic->device);
#ifdef HAVE_HDTRACE
		hdS_addValue(group, mode_name, INT32, "MODE", "EED_MODE");
#ifdef ENABLE_STATE_TRACING
		hdS_addValue(group, state_name, INT32, "MBit", "EED_MBIT");
#endif
#endif
#ifdef HAVE_DBC
		// create metric in database
		db_createNodeMetric(mode_name, hostname);
#ifdef ENABLE_STATE_TRACING
		db_createNodeMetric(state_name, hostname);
#endif
#endif
		debug(1, "Initalized tracing for Nic %d\n", id);

	} else if (IS_DISK_ID(id)) {


		Disk * disk = (Disk *) resource->device;

		gchar ** path_entries = g_strsplit(disk->device, "/", 10);
		gchar * dev = strdup(path_entries[g_strv_length (path_entries) - 1]);
		g_strfreev (path_entries);

		gchar mode_name[20 + strlen(dev)];
		gchar state_name[20 + strlen(dev)];
		g_snprintf(mode_name, 20 + strlen(dev), "EED_DISK_%s_MODE", dev);
		g_snprintf(state_name, 20 + strlen(dev), "EED_DISK_%s_STATE", dev);

#ifdef HAVE_HDTRACE
		hdS_addValue(group, mode_name, INT32, "MODE", "EED_MODE");
#ifdef ENABLE_STATE_TRACING
		hdS_addValue(group, state_name, INT32, "STATE", "EED_MBIT");
#endif
#endif
#ifdef HAVE_DBC
		// create metric in database
		db_createNodeMetric(mode_name, hostname);
#ifdef ENABLE_STATE_TRACING
		db_createNodeMetric(state_name, hostname);
#endif
#endif
		g_free(dev);

		debug(1, "Initalized tracing for Disk %d\n", id);

	} else {
		//unknown device id should not occur
		debug(1, "Unknown device id: %d\n", id);
	}

}

void resource_do_tracing(Resource * resource, hdStatsGroup *group) {

	if (resource == NULL)
	return;

	gint id = resource->id;

#ifdef HAVE_DBC
	uint64_t timestamp = getSystemTimestamp();
#endif

	gint mode = -1;
	gint state = -1;

	if (IS_CORE_ID(id)) {

		Core * core = (Core *) resource->device;

		mode = core_get_mode(core);

#ifdef ENABLE_STATE_TRACING
		state = core_get_state(core);
#endif

#ifdef HAVE_HDTRACE
		hdS_writeInt32Value(group, mode);
#ifdef ENABLE_STATE_TRACING
		hdS_writeInt32Value(group, state);
#endif
#endif
#ifdef HAVE_DBC
		gchar metric[16];
		g_snprintf(metric, 20, "eed_core%d_mode", core->id);
		db_pushNodeMetric(metric, hostname, timestamp, (double) mode);
		debug(1, "db_pushNodeMetric(%s, %s, %lu, %f);\n", metric, hostname, timestamp, (double) mode);
#ifdef ENABLE_STATE_TRACING
		g_snprintf(metric, 20, "eed_core%d_state", core->id);
		db_pushNodeMetric(metric, hostname, timestamp, state);
#endif
#endif
		debug(1, "Core %d: Mode %d - State %d\n", id, mode, state);

	} else if (IS_NIC_ID(id)) {

		Nic * nic = (Nic *) resource->device;

		mode = nic_get_mode(nic);
#ifdef ENABLE_STATE_TRACING
		state = nic_get_state(nic);
#endif
#ifdef HAVE_HDTRACE
		hdS_writeInt32Value(group, mode);
#ifdef ENABLE_STATE_TRACING
		hdS_writeInt32Value(group, state);
#endif
#endif

#ifdef HAVE_DBC
		gchar * metric = g_malloc(16 + strlen(nic->device));
		g_snprintf(metric, 15 + strlen(nic->device), "eed_nic_%s_mode", nic->device);
		db_pushNodeMetric(metric, hostname, timestamp, mode);
#ifdef ENABLE_STATE_TRACING
		g_snprintf(metric, 15 + strlen(nic->device), "eed_nic_%s_state", nic->device);
		db_pushNodeMetric(metric, hostname, timestamp, state);
#endif
		g_free(metric);
#endif

		debug(1, "Nic %d: Mode %d - State %d\n", id, mode, state);

	} else if (IS_DISK_ID(id)) {

		Disk * disk = (Disk *) resource->device;

		gchar ** path_entries = g_strsplit(disk->device, "/", 10);
		gchar * dev = strdup(path_entries[g_strv_length (path_entries) - 1]);
		g_strfreev (path_entries);

		mode = disk_get_mode(disk);
#ifdef ENABLE_STATE_TRACING
		state = disk_get_state(disk);
#endif

#ifdef HAVE_HDTRACE
		hdS_writeInt32Value(group, mode);
#ifdef ENABLE_STATE_TRACING
		hdS_writeInt32Value(group, state);
#endif
#endif

#ifdef HAVE_DBC
		gchar * metric = g_malloc(16 + strlen(dev));
		g_snprintf(metric, 20 + strlen(dev), "eed_disk_%s_mode",dev);
		db_pushNodeMetric(metric, hostname, timestamp, mode);
#ifdef ENABLE_STATE_TRACING
		g_snprintf(metric, 20 + strlen(dev), "eed_disk_%s_state", dev);
		db_pushNodeMetric(metric, hostname, timestamp, state);
#endif
		g_free(metric);
#endif

		g_free(dev);

		debug(1, "Disk %d: Mode %d - State %d\n", id, mode, state);

	} else {
		//unknown device id should not occur
		debug(1, "Unknown device id: %d\n", id);
	}
}

#endif
