/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Resource.h
 *
 *  Created on: Jul 11, 2011
 *      Author: Timo Minartz
 */

#ifndef RESOURCE_H_
#define RESOURCE_H_

#include "defines.h"

#ifdef ENABLE_TRACE
#ifdef HAVE_HDTRACE
#include "hdStats.h"
#else
#define hdStatsGroup gchar
#endif

#ifdef HAVE_DBC
static gchar hostname[8];
#endif

#endif


/**
 * A Resource stores a list of users of this resource.
 * Further mode requests are stored in an additional list.
 * To simplify operations on the lists, the size of the lists is also stored explicitly.
 */
typedef struct {
	/**
	 * ID of this Resource, to identify type (Core, Nic or Disk)
	 */
	gint id;
	/**
	 * Pointer to the Device
	 */
	gpointer device;
	/**
	 * List of users of this resource (e.g. tupel of tag and MPI rank).
	 */
	GList *users;
	/**
	 * List of requests to process for this resource
	 */
	GList *requests;
	/**
	 * Size of requests list
	 */
	gint countRequests;
	/**
	 * Size of users list
	 */
	gint countUsers;

} Resource;

/**
 * Allocates a new Resource
 * @param device_id Internal device ID (CPU_ID, NET_ID or DISK_ID)
 * @param device Pointer to the device
 * @param interval Interval in microseconds for process_list
 * @return a newly allocated Resource. Must be freed with resource_unref().
 */
Resource * resource_new(gint device_id, gpointer device);


/**
 * Clears and frees the internal lists and frees the Resource itself.
 * @param resource Resource to free
 */
void resource_unref(Resource * resource);

void resource_unref_full(Resource * resource);

/**
 * Register a new user of the Resource
 * To identify a user, a tag and the rank id is needed.
 * @param resource Resource to be used
 * @param tag Tag of the user
 * @param rank MPI rank of this user
 */
void resource_register(Resource * resource, gint tag, gint rank);

/**
 * Unregister an existing user from using the Resource.
 * To identify a user, the tag and the rank id is needed.
 * @param resource Resource to stop using
 * @param tag tag of the user
 * @param rank MPI rank of this user
 */
void resource_unregister(Resource * resource, gint tag, gint rank);

void resource_unregister_tag(Resource * resource, gint tag);

/**
 * Request a new Mode for the Resource from this users point of view.
 * Any existing requests by this user will be updated.
 * @param resource Resource to submit information for
 * @param tag of this user
 * @param rank MPI rank of this user
 * @param mode Mode to request
 */
void resource_mode_request(Resource *resource, gint tag, gint rank, Mode mode);

/**
 * When time passed, the Resource should be in the submitted mode. Until than, the mode is unchanged.
 * Any existing requests by this user will be updated.
 * @param resource Resource to submit information for
 * @param mode to to switch to after time passed
 * @param tag of this user
 * @param rank MPI rank of this user
 * @param micros Time in microseconds. When time passed, the Resource should be in mode.
 */
void resource_time_request(Resource *resource, gint mode, gint tag, gint rank, glong micros);

/**
 * Finds the minimum mode this Resource should operate in and updates requests.
 * If no user registered, the Resource is assumed to unused and MODE_UNUSED is returned.
 * If no requests exists (but registered users), MODE_MAX is returned as default value.
 * Otherwise, the time specified in rm_newModeRequest() is reduced by time_passed.
 * If the remaining idle time is smaller than twice the startup_time, MODE_MAX is returned.
 * If less requests than users are in the lists, MODE_MAX is returned.
 * In all other cases, the minimum mode is returned based on all requests in the list.
 * @param resource Resource to get Mode for
 * @param startup_time Time in microseconds the Resource needs to be in MODE_MAX again
 * @param time_passed Time in microseconds passed since the last invocation of this method
 * @return minimum Mode this Resource should operate in
 */
Mode resource_process_list(Resource *resource, glong startup_time, glong time_passed);


/**
 * Switches the Resource to the desired mode. If already in this mode, nothing happens. Otherwise the switch is initiated calling the right method based on the type of the resource
 * @param resource Resource to operate on
 * @param mode Mode to switch to
 */
void resource_switch_to_mode(Resource * resource, Mode mode);

/**
 * Returns the startup time of the Resource based on the type.
 * @param resource Resource to get startup time from
 * @return Startup time in microseconds or 0 if resource is NULL
 */
glong resource_get_startup_time(Resource * resource);

gboolean resource_is_registered_by_tag_and_rank(Resource * resource, gint tag, gint rank);

#ifdef ENABLE_TRACE

void resource_init_tracing(Resource * resource, hdStatsGroup *group);

void resource_do_tracing(Resource * resource, hdStatsGroup *group);

#endif

#endif /* RESOURCE_H_ */
