/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * ResourcesTable.c
 *
 *  Created on: Jul 12, 2011
 *      Author: Timo Minartz
 */
#include "ResourcesTable.h"
#include "Core.h"
#include "print.h"

ResourcesTable* resources_table_new() {
	ResourcesTable * rt = g_malloc(sizeof(ResourcesTable));

	rt->resources = g_hash_table_new(g_direct_hash, g_direct_equal);
	rt->registrations = 0;
	rt->constrained_resources = NULL;
	rt->processing_threads = NULL;
	rt->count_resources = 0;

	return rt;
}

void resources_table_add_constrained_resource(ResourcesTable * rt,
		ConstrainedResource * cr) {

	if (rt != NULL && cr != NULL) {
		rt->constrained_resources = g_list_prepend(rt->constrained_resources,
				cr);

		rt->count_resources += g_list_length(cr->resources);
	}
}

void resources_table_add(ResourcesTable * rt, Resource * resource) {
	if (rt != NULL) {

		// check if Resource already in table
		if(g_hash_table_lookup(rt->resources, GINT_TO_POINTER(resource->id)) != NULL)
				return;

		// add new Resource to list, key is the the resource id (e.g. 7)
		g_hash_table_insert(rt->resources, GINT_TO_POINTER(resource->id),
				(gpointer) resource);

		// add new Resource to list of all Resources with the same type

		// get list
		GList * resources_by_type_id = (GList*) g_hash_table_lookup(rt->resources,
				GINT_TO_POINTER(GET_DEV_TYPE(resource->id)));

		// add to list
		resources_by_type_id = g_list_prepend(resources_by_type_id, resource);

		// and store
		g_hash_table_insert(rt->resources,
				GINT_TO_POINTER(GET_DEV_TYPE(resource->id)),
				(gpointer) resources_by_type_id);

		rt->count_resources++;

	}
}

void resources_table_delete_resource(ResourcesTable * rt, Resource * resource) {

	if (rt != NULL && resource != NULL) {

		gboolean ret = g_hash_table_remove(rt->resources,
				GINT_TO_POINTER(resource->id));

		if (ret) {

			GList * resources_id = (GList*) g_hash_table_lookup(rt->resources,
					GINT_TO_POINTER(GET_DEV_TYPE(resource->id)));

			resources_id = g_list_remove(resources_id, (gpointer) resource);

			g_hash_table_insert(rt->resources,
					GINT_TO_POINTER(GET_DEV_TYPE(resource->id)),
					(gpointer) resources_id);

			rt->count_resources--;

		}

	}
}

void resources_table_register(ResourcesTable * rt, gint tag, gint rank,
		gint core_id) {

	if (rt != NULL) {

		GHashTableIter iter;
		gpointer key, value;

		// iterate over all resources and register corresponding ones

		g_hash_table_iter_init(&iter, rt->resources);
		while (g_hash_table_iter_next(&iter, &key, &value)) {

			gint id = GPOINTER_TO_INT(key);

			if (IS_CORE_ID(id)) {

				Resource * resource = (Resource *) value;

				Core * core = (Core *) resource->device;

				// register only if core id is the same as the submitted one

				if (core->id == core_id)
					resource_register(resource, tag, rank);

			} else if (!IS_DEV_ALL(id)) {

				// register all shared resources (nics and disks)

				Resource * resource = (Resource *) value;

				resource_register(resource, tag, rank);
			} // else ignore, it is a DEV_ALL_ID

		}

		rt->registrations++;
	}
}

void resources_table_unregister(ResourcesTable *rt, gint tag, gint rank) {
	if (rt != NULL) {

		GHashTableIter iter;
		gpointer key, value;

		g_hash_table_iter_init(&iter, rt->resources);
		while (g_hash_table_iter_next(&iter, &key, &value)) {

			gint id = GPOINTER_TO_INT(key);

			if (!IS_DEV_ALL(id)) {
				Resource * resource = (Resource *) value;

				resource_unregister(resource, tag, rank);
			}
		}
		rt->registrations--;
	}
}

void resources_table_unregister_tag(ResourcesTable *rt, gint tag) {

	if (rt != NULL) {

		GHashTableIter iter;
		gpointer key, value;

		g_hash_table_iter_init(&iter, rt->resources);
		while (g_hash_table_iter_next(&iter, &key, &value)) {

			gint id = GPOINTER_TO_INT(key);

			if (!IS_DEV_ALL(id)) {
				Resource * resource = (Resource *) value;

				resource_unregister_tag(resource, tag);
			}
		}
		rt->registrations--;
	}

}

Resource * resources_table_get_resource_by_id(ResourcesTable *rt, gint id) {
	if (rt == NULL)
		return NULL;

	return (Resource*) g_hash_table_lookup(rt->resources, GINT_TO_POINTER(id));
}

GList* resources_table_get_resources_by_id(ResourcesTable * rt, int id) {

	if (rt == NULL)
		return NULL;

	return (GList*) g_hash_table_lookup(rt->resources,
			GINT_TO_POINTER(GET_DEV_TYPE(id)));
}

GList* resources_table_get_resources_by_user_and_type(ResourcesTable *rt,
		gint tag, gint rank, gint myid) {

	GList * resources = NULL;

	if (rt != NULL) {

		GHashTableIter iter;
		gpointer key, value;

		g_hash_table_iter_init(&iter, rt->resources);
		while (g_hash_table_iter_next(&iter, &key, &value)) {

			gint id = GPOINTER_TO_INT(key);

			if (!IS_DEV_ALL(id)) {
				Resource * resource = (Resource *) value;

				if (ID_IS_MINE(id, myid) && resource_is_registered_by_tag_and_rank(resource, tag, rank))
					resources = g_list_prepend(resources, resource);
			}
		}
	}

	return resources;
}

GList* resources_table_get_resources_by_user(ResourcesTable *rt, gint tag,
		gint rank) {

	GList * resources = NULL;

	if (rt != NULL) {

		GHashTableIter iter;
		gpointer key, value;

		g_hash_table_iter_init(&iter, rt->resources);
		while (g_hash_table_iter_next(&iter, &key, &value)) {

			gint id = GPOINTER_TO_INT(key);

			if (!IS_DEV_ALL(id)) {
				Resource * resource = (Resource *) value;

				if (resource_is_registered_by_tag_and_rank(resource, tag, rank)) {
					resources = g_list_prepend(resources, resource);
				}
			}
		}
	}

	return resources;
}

gint resources_table_get_count_registrations(ResourcesTable * rt) {
	if (rt == NULL)
		return -1;

	return rt->registrations;
}

void unref_table(ResourcesTable * rt, gboolean unref_full) {

	if (rt != NULL) {

		GList * cores = g_hash_table_lookup(rt->resources,
				GINT_TO_POINTER(CORE_ALL_ID));
		if (cores != NULL) {
			g_list_free(cores);
		}

		GList * nics = g_hash_table_lookup(rt->resources,
				GINT_TO_POINTER(NIC_ALL_ID));
		if (nics != NULL) {
			g_list_free(nics);
		}

		GList * disks = g_hash_table_lookup(rt->resources,
				GINT_TO_POINTER(DISK_ALL_ID));
		if (disks != NULL) {
			g_list_free(disks);
		}

		if (unref_full) {

			// free all ConstrainedResources

			for (GList * it = g_list_first(rt->constrained_resources);
					it != NULL; it = it->next) {
				constrained_resource_unref((ConstrainedResource *) it->data);
			}
			g_list_free(rt->constrained_resources);

			// free all Resources

			GHashTableIter iter;
			gpointer key, value;

			g_hash_table_iter_init(&iter, rt->resources);
			while (g_hash_table_iter_next(&iter, &key, &value)) {

				gint id = GPOINTER_TO_INT(key);

				if (!IS_DEV_ALL(id)) {
					Resource * resource = (Resource *) value;
					resource_unref_full(resource);
				}
			}
		}

		g_hash_table_remove_all(rt->resources);
		g_hash_table_destroy(rt->resources);

		g_free(rt);
	}
}

void resources_table_unref(ResourcesTable * rt) {
	unref_table(rt, FALSE);
}

void resources_table_unref_full(ResourcesTable * rt) {
	unref_table(rt, TRUE);
}

void resources_table_start_process_resources(ResourcesTable *rt) {

	if (rt != NULL) {
		rt->processing_threads = g_malloc(
				g_list_length(rt->constrained_resources) * sizeof(GThread*));

		if (!g_thread_get_initialized())
			g_thread_init(NULL);

		gint i = 0;
		for (GList * it = g_list_first(rt->constrained_resources); it != NULL;
				it = it->next) {

			// start thread that processes this ConstrainedResource

			rt->processing_threads[i++] = g_thread_create(
					(GThreadFunc) &constrained_resource_process, it->data, TRUE,
					NULL);

			debug(1, "Started processing thread %d...\n", i-1);
		}
	}
}

void resources_table_stop_process_resources(ResourcesTable *rt) {

	if (rt == NULL)
		return;

	for (GList * it = g_list_first(rt->constrained_resources); it != NULL; it =
			it->next) {
		ConstrainedResource * cr = (ConstrainedResource *) it->data;

		cr->finalize = TRUE;
	}

	gint i = 0;

	for (GList * it = g_list_first(rt->constrained_resources); it != NULL; it =
			it->next) {
		debug(1, "Waiting for processing thread %d...\n", i);
		g_thread_join(rt->processing_threads[i++]);
		debug(1, "Waiting for processing thread %d...Done\n", i-1);
	}

	g_free(rt->processing_threads);
	rt->processing_threads = NULL;
}

gint resources_table_get_count_resources(ResourcesTable *rt) {
	return rt->count_resources;
}

#ifdef ENABLE_TRACE
void resources_table_init_tracing(ResourcesTable * rt, hdStatsGroup * group) {

	if (rt == NULL)
	return;

	GHashTableIter iter;
	gpointer key, value;

	g_hash_table_iter_init(&iter, rt->resources);
	while (g_hash_table_iter_next(&iter, &key, &value)) {

		gint id = GPOINTER_TO_INT(key);

		if (!IS_DEV_ALL(id)) {
			Resource * resource = (Resource *) value;

			resource_init_tracing(resource, group);
		}
	}
}
void resources_table_do_tracing(ResourcesTable * rt, hdStatsGroup * group) {

	if (rt == NULL)
	return;

	GHashTableIter iter;
	gpointer key, value;

	g_hash_table_iter_init(&iter, rt->resources);
	while (g_hash_table_iter_next(&iter, &key, &value)) {

		gint id = GPOINTER_TO_INT(key);

		if (!IS_DEV_ALL(id)) {
			Resource * resource = (Resource *) value;

			resource_do_tracing(resource, group);
		}
	}
}
#endif

