/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * Server.c
 *
 *  Created on: Dec 6, 2010
 *      Author: Timo Minartz
 */
#include <glib.h>
#include <glib-object.h>
#include <gio/gio.h>
#include <signal.h> /* signals */
#include <string.h>

#include "Server.h"
#include "Command.h"
#include "print.h"
#include "Daemon.h" /* for initializing mutex */

static Server * server = NULL;

void server_stop() {

	if (server != NULL) {

		/* set boolean for terminating server thread */
		server->terminate_server = TRUE;
		if (server->cancellable != NULL) {
			g_cancellable_cancel(server->cancellable);
		}
	}
}

/**
 * Handels signal for SIGTERM and SIGINT
 */
void sighandler(gint sig) {

	if (sig == SIGTERM)
		info("Received termination signal.\n");

	if (sig == SIGINT)
		info("Received interrupt from keyboard.\n");

	server_stop();

}

void handle_connection(GSocketConnection * connection) {

	if (server == NULL)
		return;

	info("Accepted connection on port %d.\n", server->port);

	/* open function module in main context */

	GModule *module = g_module_open(NULL, G_MODULE_BIND_LAZY);

	if (module == NULL) {
		error("Cannot open module main. Abort handling connection.\n");
		return;
	}

	GInputStream * stream = g_io_stream_get_input_stream(
			(GIOStream*) connection);

	while (!server->terminate_server) {

		debug(1, "Waiting for command...\n");
		Command * command;

		gboolean ret = command_receive(stream, server->cancellable, &command);

		if (!ret) {
			debug(1, "Command receive cancelled.\n");
			break;
		}

		gchar * cmd_string = command_get_string(command);

		debug(3, "Received command %s.\n", cmd_string);

		// because we are in the main context, we have to check for valid prefix
		// (to avoid remote calling of any method)
		if (!g_str_has_prefix(cmd_string, "ee_")) {
			g_warning("No valid prefix for command %s. Will be ignored.\n",
					command_get_name(command));
			command_unref(command);
			g_free(cmd_string);
			continue;
		}

		// handle command

		if (!command_execute(command, module)) {
			// no function with corresponding name and signature available
			info("No function \"void %s\" available in module \"%s\". Ignoring this message.\n",cmd_string, g_module_name(module));
		} else {
			debug(1, "Executed command %s.\n", cmd_string);
		}

		command_unref(command);

		if (g_str_has_prefix(cmd_string, "ee_finalize")) {
			// terminate this connection
			g_free(cmd_string);
			break;
		}

		g_free(cmd_string);

	}

	// results in GLib-GIO-CRITICAL **: g_output_stream_close: assertion `G_IS_OUTPUT_STREAM (stream)' failed
	//	g_object_unref(stream); GLib Documentation says: Do not free....

	g_object_unref(connection);

	// it is not really the remaining number of open connections (because threads are shut down after they're idle for a defined time...)
	//GLib-CRITICAL **: g_thread_pool_get_num_threads: assertion `real->running' failed
	//info("Closed connection. Remaining open connections: %d\n", g_thread_pool_get_num_threads(server->workerPool) - 1);

	if (!g_module_close(module))
		error("Error while trying to close module: %s", g_module_error());

	info("Closed connection on port %d.\n", server->port);

}

Server * server_new(gint socket_port) {

	if (server == NULL) {
		/* register termination signal */
		signal(SIGTERM, sighandler);
		/* register Interrupt from keyboard for ctrl-c */
		signal(SIGINT, sighandler);

		/* threadsafe glib */
		if (!g_thread_get_initialized())
			g_thread_init(NULL);

		g_type_init();

		server = g_malloc(sizeof(Server));

		server->terminate_server = FALSE;
		server->port = socket_port;
		server->instances = 0;

		// create a worker pool for processing the connections
		GError * error = NULL;
		server->workerPool = g_thread_pool_new((GFunc) &handle_connection,
				NULL, -1, FALSE, &error);

		// create an cancellable object to be able to cancel the socket_listener_accept
		server->cancellable = g_cancellable_new();

	} else {
		if (server->port != socket_port) {
			error("There is already a server running on port %d. Abort.", server->port);
			return NULL;
		}
	}

	server->instances++;

	return server;
}

gboolean server_run() {

	if (server != NULL) {

		registerThread(g_thread_self(), "server");

		/* init gio connection */

		GSocketListener * listen;
		GSocketConnection * conn;
		GError * error = NULL;

		g_type_init();

		listen = g_socket_listener_new();
		g_socket_listener_add_inet_port(listen, server->port, NULL, NULL);

		while (1) {

			info("Waiting for connection on port %d...\n", server->port);

			conn = g_socket_listener_accept(listen, NULL, server->cancellable,
					&error);

			if (error != NULL && error->code == G_IO_ERROR_CANCELLED) {
				debug(1, "Canceled listener accept.\n");
				if (conn != NULL)
					g_object_unref(conn);

				g_error_free(error);
				break;
			}

			error = NULL;
			g_thread_pool_push(server->workerPool, conn, &error);
			if (error != NULL) {
				error("Creating new thread from pool to process connection failed with message: %s\n", error->message);
				if (conn != NULL)
					g_object_unref(conn);
				g_error_free(error);
				break;
			}

			if (server->terminate_server)
				break;
		}

		g_object_unref(listen);

		return TRUE;

	}

	return FALSE;
}

void server_unref(Server * instance) {

	if (instance == NULL) {
		g_warning("Trying to unref NULL pointer.\n");
		return;
	}

	if (instance != server) {
		g_warning("Trying to unref wrong object type.\n");
		return;
	}

	if (server != NULL) {

		server->instances--;

		if (server->instances == 0) {

			server_stop();

			/* waits until all threads inside of the pool finishes its work */
			g_thread_pool_free(server->workerPool, FALSE, TRUE);

			g_object_unref(server->cancellable);

			g_free(server);

			server = NULL;
		}
	}
}
