/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * ConfigTest.c
 *
 *  Created on: Jul 19, 2011
 *  Author: Timo Minartz
 */

#include "ConfigTest.h"
#include "TestMethods.h"
#include "Config.h"
#include "ConstrainedResource.h"
#include <unistd.h> /* for sleep */

void test_config_create_resources_table(void) {

	Config * config;

	ResourcesTable * rt;

	GList * crs;

	if (is_test_intel_node()) {

		config = config_create_Intel();

		g_assert_cmpint(config->frequency_per_core, ==, FALSE);

		rt = config_create_resources_table(config);

		crs = rt->constrained_resources;

		// socket0, socket1, nic, disk
		gint count_devices = 4;

		// disk is only allocated if in root_mode
		if (!is_test_root_context())
			count_devices = 3;

		g_assert_cmpint(g_list_length(crs), ==, count_devices);

		// remember if each device is checked...
		gboolean reached[count_devices];

		for (gint i = 0; i < count_devices; ++i) {
			reached[i] = FALSE;
		}

		for (GList * it = g_list_first(crs); it != NULL; it = it->next) {

			ConstrainedResource * cr = (ConstrainedResource *) it->data;

			if (g_list_length(cr->resources) == 1) {
				// nic or disk

				gint id = ((Resource *) cr->resources->data)->id;

				if (IS_DISK_ID(id)) {
					g_assert_cmpint(id, ==, 201);
					reached[3] = TRUE;
				} else if (IS_NIC_ID(id)) {
					g_assert_cmpint(id, ==, 101);
					reached[2] = TRUE;
				} else {
					g_assert_not_reached();
				}

			} else {

				//TODO: fix for smt enabled
				g_assert_cmpint(g_list_length(cr->resources), ==, 4);

				Resource * resource = resources_table_get_resource_by_id(rt, 1);

				if (g_list_find(cr->resources, resource) != NULL) {

					// socket 0

					for (gint id = 1; id < 5; ++id) {
						resource = resources_table_get_resource_by_id(rt, id);
						g_assert(g_list_find(cr->resources, resource) != NULL);
					}

					reached[0] = TRUE;

				} else {

					// socket 1

					for (gint id = 5; id < 9; ++id) {
						resource = resources_table_get_resource_by_id(rt, id);
						g_assert(g_list_find(cr->resources, resource) != NULL);
					}

					reached[1] = TRUE;
				}

			}

		}

		for (gint i = 0; i < count_devices; ++i) {
			g_assert_cmpint(reached[i], ==, TRUE);
		}
	} else if (is_test_amd_node()) {

		config = config_create_AMD();

		g_assert_cmpint(config->frequency_per_core, ==, TRUE);

		rt = config_create_resources_table(config);

		crs = rt->constrained_resources;

		// 24 cores + nic + disk

		gint count_devices = 26;

		// disk is only allocated if in root_mode
		if (!is_test_root_context())
			count_devices = 25;

		g_assert_cmpint(g_list_length(crs), ==, count_devices);

		// remember if each device is checked...
		gboolean reached[count_devices];

		for (gint i = 0; i < count_devices; ++i) {
			reached[i] = FALSE;
		}

		for (GList * it = g_list_first(crs); it != NULL; it = it->next) {

			ConstrainedResource * cr = (ConstrainedResource *) it->data;

			// one constrained resource for each resource
			g_assert_cmpint(g_list_length(cr->resources), ==, 1);

			gint id = ((Resource *) cr->resources->data)->id;

			if (IS_CORE_ID(id)) {
				reached[id - 1] = TRUE;
			} else if (IS_DISK_ID(id)) {
				g_assert_cmpint(id, ==, 201);
				reached[25] = TRUE;
			} else if (IS_NIC_ID(id)) {
				g_assert_cmpint(id, ==, 101);
				reached[24] = TRUE;
			} else {
				g_assert_not_reached();
			}

		}

		for (gint i = 0; i < count_devices; ++i) {
			g_assert_cmpint(reached[i], ==, TRUE);
		}
	}

	resources_table_unref_full(rt);

	config_unref(config);

	// wait for all possible state transitions...

	sleep(5);

}

void test_config_read_write(void) {


	GHashTable * socketMapping =
			g_hash_table_new(g_direct_hash, g_direct_equal);
	for (gint i = 0; i < 4; ++i)
		g_hash_table_insert(socketMapping, GINT_TO_POINTER(i), GINT_TO_POINTER(0));
	for (gint i = 4; i < 8; ++i)
			g_hash_table_insert(socketMapping, GINT_TO_POINTER(i), GINT_TO_POINTER(1));

	GList * disks = NULL;
	disks = g_list_prepend(disks, "/dev/sda");
	GList * nics = NULL;
	nics = g_list_prepend(nics, "eth0");
	glong *startup_times = g_malloc(3 * sizeof(glong));
	for (gint i = 0; i < 3; ++i)
		startup_times[i] = 100;

	Config * in = config_new(2, socketMapping, disks, nics, startup_times,
			FALSE, 2000);

	g_assert_cmpint(in->countSockets, ==, 2);

	Config * out = NULL;

	g_print("Writing test config to /tmp/config.\n");

	g_assert_cmpint(config_write(in, "/tmp/config"), ==, TRUE);

	g_print("Read test config from /tmp/config.\n");

	g_assert_cmpint(config_read(&out, "/tmp/config"), ==, TRUE);

	g_assert_cmpint(in->countSockets, ==, out->countSockets);
	g_assert_cmpint(in->frequency_per_core, ==, out->frequency_per_core);
	g_assert_cmpint(in->interval, ==, out->interval);

	GHashTableIter iter;
	gpointer key, value;

	// check for correct socket mapping

	g_hash_table_iter_init(&iter, out->socketMapping);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		if (GPOINTER_TO_INT(key) >= 0 && GPOINTER_TO_INT(key) <= 3)
			g_assert_cmpint(0, ==, GPOINTER_TO_INT(g_hash_table_lookup(out->socketMapping, key)));
		else if (GPOINTER_TO_INT(key) >= 4 && GPOINTER_TO_INT(key) <= 7)
			g_assert_cmpint(1, ==, GPOINTER_TO_INT(g_hash_table_lookup(out->socketMapping, key)));
		else
			g_assert_not_reached();
	}

	g_hash_table_iter_init(&iter, in->socketMapping);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		g_assert_cmpint(GPOINTER_TO_INT(value), ==, GPOINTER_TO_INT(g_hash_table_lookup(out->socketMapping, key)));
	}

	// check for correct device mapping

	g_hash_table_iter_init(&iter, in->deviceMapping);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		gint id = GPOINTER_TO_INT(key);

		if (IS_CORE_ID(id))
			g_assert_cmpint(GPOINTER_TO_INT(value), ==, GPOINTER_TO_INT(g_hash_table_lookup(out->deviceMapping, key)));
		else
			g_assert_cmpstr((gchar*) value, ==, (gchar*) g_hash_table_lookup(out->deviceMapping, key));
	}

	// free the lists
	g_list_free(disks);
	g_list_free(nics);

	// free the Configs

	config_unref(in);
	config_unref(out);
}
