/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * DaemonTest.c
 *
 *  Created on: Jul 12, 2011
 *      Author: Timo Minartz
 */
#include "DaemonTest.h"
#include "Daemon.h"
#include "defines.h"
#include "Config.h"
#include "Core.h"
#include "TestMethods.h"
#include <unistd.h>		/* needed for gethostname */
#include <stdio.h> 		/* needed for remove */
#include <stdlib.h> /* needed for setenv */
#include <glib.h>

#ifdef HAVE_DBC
#include "DBConnector.h"
#include "MetricInterpreter.h"
#else

#include <sys/time.h>
#include <stdint.h>

uint64_t getSystemTimestamp() {
	struct timeval detail_time;
	gettimeofday(&detail_time, NULL);

	uint64_t microseconds = detail_time.tv_sec * 1000000 + detail_time.tv_usec;

	return microseconds;

}

#endif

void test_daemon_init(void) {

	gint tag = 1313;
	gint rank = 0;
	gint core_id = 0;

	ee_init(tag, rank, "test_server_eed", core_id);

	sleep(5);

	/*
	 * TODO: check progress:
	 * device initiation -> MODE_DEFAULT
	 * no user registered -> MODE_UNUSED
	 * user registered, but no mode change -> MODE_MAX
	 * users unregistered -> MODE_UNUSED
	 * device unref -> MODE_DEFAULT
	 */

	ee_finalize(tag, rank);

	sleep(5);

}

void test_daemon_dev_mode(void) {

	/**
	 * Register one process using the ee_dev_mode interface
	 */

	gint tag = 1313;
	gint rank = 0;
	glong interval = 10000; // 10000 microseconds = 10 milliseconds
	glong max_wait = 10000000; // 10000000 microseconds = 10 seconds
	glong wait = 0;
	gint core_id = 0;

	ee_init(tag, rank, "test_server_eed", core_id);

	// Now devices should be initiated. We initialize one control device, which should point to the core wit id 0
	Core * core = core_new(0);

	ee_dev_mode(CORE_ALL_ID, MODE_MAX, tag, rank);

	g_usleep(1000000);

	if (core->has_turboboost) {

		wait = 0;
		while (core_get_state(core) != core->frequencies[1] && wait < max_wait) {
			g_usleep(interval);
			wait += interval;
		}

		g_print("Core reached MODE_READY state after %lu microseconds.\n", wait);
		g_assert_cmpint(core_get_state(core), ==, core->frequencies[1]);
	} else {

		wait = 0;
		while (core_get_state(core) != core->frequencies[0] && wait < max_wait) {
			g_usleep(interval);
			wait += interval;
		}

		g_print("Core reached MODE_READY state after %lu microseconds.\n", wait);

		g_assert_cmpint(core_get_state(core), ==, core->frequencies[0]);
	}

	ee_dev_mode(CORE_ALL_ID, MODE_UNUSED, tag, rank);

	wait = 0;
	while (core_get_state(core) != core->frequencies[core->count_frequencies
			- 1] && wait < max_wait) {
		g_usleep(interval);
		wait += interval;
	}

	g_print("Core reached MODE_UNUSED state after %lu microseconds.\n", wait);

	g_assert_cmpint(core_get_state(core), ==, core->frequencies[core->count_frequencies - 1]);

	ee_dev_mode(CORE_ALL_ID, MODE_MIN, tag, rank);

	wait = 0;
	while (core_get_state(core) != core->frequencies[core->count_frequencies
			- 1] && wait < max_wait) {
		g_usleep(interval);
		wait += interval;
	}

	g_print("Core reached MODE_MIN state after %lu microseconds.\n", wait);

	g_assert_cmpint(core_get_state(core), ==, core->frequencies[core->count_frequencies - 1]);

	ee_finalize(tag, rank);

	core_unref(core);

}

void test_daemon_rms(void) {
}

#ifdef ENABLE_TRACE
void test_daemon_tracing(void) {

	gchar * prgname = "test_tracing";

	gchar host [8];
	gethostname(host, 8);

	gint tag = 1313;

	gint rank = 0;
	gint sleeptime = 10000000; // 10 secs

	gchar verbstr[12];

	gint intervals[] = {1000000, 100000, 10000, 1000, 0};

	// min tracing interval is somewhere at 300 microseconds, since one tracing step needs approximately 271 microseconds

	for(gint i=0; i<5; ++i) {

		gint interval = intervals[i];

		snprintf (verbstr, 11, "%d", interval);
		setenv("EED_TRACE_INTERVAL", verbstr, 1);

#ifdef HAVE_HDTRACE
		/* delete tracing file if it already exists */

		GString * tracefilename = g_string_new("");

		g_string_printf(tracefilename, "/tmp/%s_%s_eed.stat", prgname, host);

		remove(tracefilename->str);

#endif

		// get current timestamp
		uint64_t timestamp_before = getSystemTimestamp();

		/* calling ee_init which should initialize the tracing and start the tracing thread */

		ee_init(tag, rank, prgname, 0);
		uint64_t timestamp_after_init = getSystemTimestamp();

		g_usleep(sleeptime); // sleep to get some tracing results...

		/* calling ee_finalize should finalize the tracing and join the tracing thread */

		uint64_t timestamp_before_finalize = getSystemTimestamp();

		ee_finalize(tag, rank);

		// get current timestamp
		uint64_t timestamp_after = getSystemTimestamp();

		g_print("ee_init: %ld microseconds. ee_finalize: %ld microseconds. total: %ld microseconds\n", timestamp_after_init - timestamp_before, timestamp_after - timestamp_before_finalize, timestamp_after - timestamp_before);

#ifdef HAVE_DBC
		// get values from database

		db_initConnection(1);
		row_struct * rows;

		gchar * mode = "eed_core0_mode";
		gint ret1 = db_getNodeMetric(mode, host, timestamp_before, timestamp_after, &rows);

		g_assert_cmpint(ret1, >=, 0);

		if(ret1 > 0)
		g_free(rows);

		if(interval > 1000) {
			// check if there are enough values in db ( >= 100% of the theoretical count)
			g_assert_cmpint(ret1, >=, (int) (mi_getTimestampInSec(sleeptime) / mi_getTimestampInSec(interval)));
		} else if (interval > 0) {
			// check if there are enough values in db ( >= 90% of the theoretical count)
			g_assert_cmpint(ret1, >=, (int) (mi_getTimestampInSec(sleeptime) / mi_getTimestampInSec(interval) * 0.90));
		} else {
			g_assert_cmpint(ret1, ==, 0);
		}

#ifdef ENABLE_STATE_TRACING
		gchar * state = "eed_core0_state";
		gint ret2 = db_getNodeMetric(state, host, timestamp_before, timestamp_after, &rows);

		g_assert_cmpint(ret1, ==, ret2);

		g_free(rows);
#endif

		db_finishConnection();

#endif

#ifdef HAVE_HDTRACE

		/* now open the trace file */

		gchar * buffer;
		gsize length;
		GError *error = NULL;

		if(interval > 0) {
			if (!g_file_get_contents(tracefilename->str, &buffer, &length, &error)) {
				g_assert_not_reached();
			}

			// cannot check length: is variable based on do_tracing calls....
			g_assert_cmpint(length, >, 0);

			//TODO: more checks here... but needs an trace reader in c!

			g_assert_cmpint(remove(tracefilename->str), ==, 0);

			g_string_free(tracefilename, TRUE);

			g_free(buffer);
		} else {
			if (g_file_get_contents(tracefilename->str, &buffer, &length, &error)) {
				g_assert_not_reached();
			}
		}

#endif
	}

}
#endif
