/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * ResourceTest.c
 *
 *  Created on: Jul 11, 2011
 *      Author: Timo Minartz
 */

#include <glib.h>

#include "ResourceTest.h"
#include "Resource.h"
#include "Core.h"
#include "Config.h"

#ifdef HAVE_HDTRACE
#include <unistd.h>		/* needed for gethostname */
#include <stdio.h> 		/* needed for remove */
#include "hdTopo.h"
#include "hdStats.h"
#include "Nic.h"
#include "Disk.h"
#include "TestMethods.h"
#endif

void test_resource_register() {
	gint tag1 = 1313;
	gint tag2 = 1413;

	gint rank1 = 1;
	gint rank2 = 2;
	gint rank3 = 3;

	Core * core1 = core_new(1);
	Core * core2 = core_new(2);

	Resource * resource1 = resource_new(1, core1);
	Resource * resource2 = resource_new(2, core2);

	if(resource1 == NULL || resource2 == NULL) {
		// no tests possible
		return;
	}

	g_assert_cmpint(resource1->countUsers, ==, 0);
	g_assert_cmpint(resource2->countUsers, ==, 0);

	resource_register(resource1, tag1, rank1);
	resource_register(resource2, tag2, rank2);

	g_assert_cmpint(resource1->countUsers, ==, 1);
	g_assert_cmpint(resource2->countUsers, ==, 1);

	resource_register(resource1, tag2, rank1);

	g_assert_cmpint(resource1->countUsers, ==, 2);
	g_assert_cmpint(resource2->countUsers, ==, 1);

	resource_register(resource2, tag2, rank3);

	g_assert_cmpint(resource1->countUsers, ==, 2);
	g_assert_cmpint(resource2->countUsers, ==, 2);

	resource_unregister_tag(resource2, tag2);

	g_assert_cmpint(resource1->countUsers, ==, 2);
	g_assert_cmpint(resource2->countUsers, ==, 0);

	resource_unregister(resource1, tag1, rank1);

	g_assert_cmpint(resource1->countUsers, ==, 1);
	g_assert_cmpint(resource2->countUsers, ==, 0);

	resource_unregister_tag(resource1, tag2);

	g_assert_cmpint(resource1->countUsers, ==, 0);
	g_assert_cmpint(resource2->countUsers, ==, 0);

	resource_unref(resource1);
	resource_unref(resource2);

	core_unref(core1);
	core_unref(core2);

}

void test_resource_process_list() {

	gint tag = 1313;
	gint rank = 0;

	Core * core = core_new(0);

	Resource * resource = resource_new(1, core);

	Mode mode = resource_process_list(resource, 200, 0);

	// no process registered -> MODE_UNUSED must be returned
	g_assert_cmpint(mode, ==, MODE_UNUSED);

	resource_register(resource, tag, rank);

	mode = resource_process_list(resource, 200, 0);

	// request list is empty -> MODE_MAX must be returned
	g_assert_cmpint(mode, ==, MODE_MAX);

	resource_mode_request(resource, tag, rank, MODE_MAX);

	mode = resource_process_list(resource, 200, 0);

	// ony one user who wants MODE_MAX -> MODE_MAX must be returned
	g_assert_cmpint(mode, ==, MODE_MAX);

	resource_mode_request(resource, tag, rank, MODE_UNUSED);

	mode = resource_process_list(resource, 200, 0);

	// ony one user who wants MODE_UNUSED -> MODE_UNUSED must be returned
	g_assert_cmpint(mode, ==, MODE_UNUSED);

	resource_mode_request(resource, tag, rank, MODE_MIN);

	mode = resource_process_list(resource, 200, 0);

	g_assert_cmpint(mode, ==, MODE_MIN);

	resource_time_request(resource, MODE_MAX, tag, rank, 10000);

	mode = resource_process_list(resource, 200, 0);

	g_assert_cmpint(mode, ==, MODE_UNUSED);

	mode = resource_process_list(resource, 200, 1000);

	g_assert_cmpint(mode, ==, MODE_UNUSED);

	mode = resource_process_list(resource, 200, 8900);

	g_assert_cmpint(mode, ==, MODE_MAX);

	mode = resource_process_list(resource, 200, 100);

	g_assert_cmpint(mode, ==, MODE_MAX);

	resource_register(resource, tag, 1);

	mode = resource_process_list(resource, 200, 0);

	g_assert_cmpint(mode, ==, MODE_MAX);

	resource_mode_request(resource, tag, rank, MODE_MIN);

	mode = resource_process_list(resource, 200, 0);

	// 2 resources in the list -> rank requests MODE_MIN , 1 doesn't request anything => MODE_MIN

	g_assert_cmpint(mode, ==, MODE_MIN);

	resource_time_request(resource, MODE_MAX, tag, 1, 10000);

	mode = resource_process_list(resource, 200, 0);

	g_assert_cmpint(mode, ==, MODE_MIN);

	mode = resource_process_list(resource, 200, 1000);

	g_assert_cmpint(mode, ==, MODE_MIN);

	resource_mode_request(resource, tag, rank, MODE_UNUSED);

	mode = resource_process_list(resource, 200, 1000);

	g_assert_cmpint(mode, ==, MODE_UNUSED);

	mode = resource_process_list(resource, 200, 7900);

	// rank 0 requests MODE_UNUSED, rank 1 requests MODE_MAX from TIME_REQUEST

	g_assert_cmpint(mode, ==, MODE_MAX);

	resource_mode_request(resource, tag, 0, MODE_MIN);

	mode = resource_process_list(resource, 200, 0);

	// rank 1 requests MODE_MAX resulting from TIME_REQUEST, rank 0 requests MODE_MIN

	g_assert_cmpint(mode, ==, MODE_MAX);

	resource_unregister(resource, tag, 1);

	resource_unref(resource);

	core_unref(core);

}

#ifdef ENABLE_TRACE
#ifdef HAVE_HDTRACE
void test_resource_tracing(void) {

	gchar * prgname = "testresourcetracing";

	gchar host [8];
	gethostname(host, 8);

	/* delete tracing file if it already exists */

	GString * tracefilename = g_string_new("");

	g_string_printf(tracefilename, "/tmp/%s_%s_eed.stat", prgname, host);

	remove(tracefilename->str);

	/* create hdstats group */

	const gchar *levels[] = {"Host"};

	hdTopology *myTopology = hdT_createTopology(prgname, levels, 1);

	const gchar *path[] = {host};
	hdTopoNode *myTopoNode = hdT_createTopoNode(myTopology, path, 1);

	hdStatsGroup *group = group = hdS_createGroupWithFilenamePrefixAndAbortIfExists("eed", myTopoNode, 1,"/tmp/",0);

	/* create test resources from each type */

	Core * core = core_new(0);
	Nic * nic = nic_new("eth0", 2000);
	Disk * disk = disk_new("/dev/sda", 2000);

	Resource * core_resource = resource_new(1, core);
	Resource * nic_resource = resource_new(101, nic);
	Resource * disk_resource = resource_new(201, disk);

	/* initialize tracing */

	resource_init_tracing(core_resource, group);
	resource_init_tracing(nic_resource, group);
	resource_init_tracing(disk_resource, group);

	hdS_commitGroup(group);

	hdS_enableGroup(group);

	/* do a tracing step -> all resources in MODE_DEFAULT */

	resource_do_tracing(core_resource, group);
	resource_do_tracing(nic_resource, group);
	resource_do_tracing(disk_resource, group);

	/* switch all resources to different modes */

	for(Mode mode = 2; mode < COUNT_MODES; ++mode) {

		core_switch_to_mode(core, mode);
		nic_switch_to_mode(nic, mode);
		disk_switch_to_mode(disk, mode);

		// wait for all mode changes to complete

		g_usleep(5000000); // 5 seconds


		// do a tracing step

		resource_do_tracing(core_resource, group);
		resource_do_tracing(nic_resource, group);
		resource_do_tracing(disk_resource, group);

	}

	/* and finalize tracing */

	hdS_disableGroup(group);

	hdS_finalize(group);

	hdT_destroyTopoNode(myTopoNode);
	hdT_destroyTopology(myTopology);

	/* unref instances */

	resource_unref_full(core_resource);
	resource_unref_full(nic_resource);
	resource_unref_full(disk_resource);

	/* now open the trace file */

	gchar * buffer;
	gsize length;
	GError *error = NULL;

	if (!g_file_get_contents(tracefilename->str, &buffer, &length, &error)) {
		g_assert_not_reached();
	}

	if(is_test_intel_node())
		g_assert_cmpint(length, ==, 849);
	else if(is_test_amd_node())
		g_assert_cmpint(length, ==, 847);

	//TODO: more checks here... but needs an trace reader in c!

	g_assert_cmpint(remove(tracefilename->str), ==, 0);

	g_string_free(tracefilename, TRUE);

	g_free(buffer);

}
#endif
#endif
