/*
 * eeDaemon
 * http://eeclust.de
 *
 * Copyright (c) 2010-2012, University of Hamburg
 *
 * See the file COPYING in the package base directory for details
 *
 * TestSuite.c
 *
 *  Created on: Jul 6, 2011
 *      Author: Timo Minartz
 */

#include <glib.h>
#include <gio/gio.h>

#include "TestMethods.h"
#include "print.h"
#include "CoreTest.h"
#include "NicTest.h"
#include "DiskTest.h"
#include "CommandTest.h"
#include "ResourceTest.h"
#include "ConstrainedResourceTest.h"
#include "ResourcesTableTest.h"
#include "ServerTest.h"
#include "DaemonTest.h"
#include "ConfigTest.h"
#include "Config.h"
#include "defines.h"
#include <unistd.h> /* for gethostname */

gchar * config_file_path = NULL;

gint main(gint argc, gchar ** argv) {

	print_debug_messages = 1;
	verbose = TRUE;

	debug(1, "Activated debug messages on level 1.\n");
	print("Be verbose.\n");

	// initialize test program
	g_test_init(&argc, &argv, NULL);

	g_type_init();

	// set path to test config file
	config_file_path = g_strdup("/tmp/config");

	// generate test config file if not existing
	if (g_file_test(DEFAULT_CONFIG_FILE, G_FILE_TEST_EXISTS)) {
		// copy the system config file to the test file
		print("Copying system config file /etc/eedaemon.conf to test config file\n");

		gchar * contents;
		gsize length = 0;
		gboolean ret = g_file_get_contents(DEFAULT_CONFIG_FILE, &contents,
				&length, NULL);
		if (!ret) {
			error("Error reading config data from config file /etc/eedaemon.conf\n");
			return 1;
		}
		ret = g_file_set_contents(config_file_path, contents, length, NULL);

		g_free(contents);

		if (!ret) {
			error("Error writing config data from config file /etc/eedaemon.conf to %s\n", config_file_path);
			return 2;
		}
	} else {
		// generate the new file based on the hostname

		gchar host[20];
		gethostname(host, 20);

		print("Generating test config file %s for host %s\n", config_file_path, host);

		Config * config = NULL;

		if (g_str_has_prefix(host, "intel")) {
			config = config_create_Intel();
		} else if (g_str_has_prefix(host, "amd")) {
			config = config_create_AMD();
		} else {
			config = config_create_testConfig();
		}

		gboolean ret = config_write(config, config_file_path);

		if (config == NULL || !ret) {
			error("Error generating config file to test config file %s\n", config_file_path);
			return 3;
		}

		config_unref(config);
	}

	// hook up test functions

	if (is_test_root_context()) {
		// need root privileges..
		g_test_add_func("/Core/state_switch", test_core_state_switch);
		g_test_add_func("/ConstrainedResource/cpu",
				test_constrained_resource_cpu);
		g_test_add_func("/Command/execute", test_command_execute);

		if (is_test_eeclust_node()) {
			// should only be executed on eeclust nodes...
			//g_test_add_func("/Nic/state_switch", test_nic_state_switch); // disabled for the moment
			g_test_add_func("/Disk/state_switch", test_disk_state_switch);
			g_test_add_func("/Daemon/daemon_dev_mode", test_daemon_dev_mode);

#ifdef ENABLE_TRACE
#ifdef HAVE_HDTRACE
			g_test_add_func("/Resource/tracing", test_resource_tracing);
#endif
#endif
		}

	}

	g_test_add_func("/Config/read_write", test_config_read_write);
	g_test_add_func("/Command/get_name", test_command_get_name);
	g_test_add_func("/Command/get_string", test_command_get_string);
	g_test_add_func("/Command/communication", test_command_communication);
	g_test_add_func("/Command/multiple_clients", test_command_multiple_clients);

	g_test_add_func("/Resource/register", test_resource_register);
	g_test_add_func("/Resource/process_list", test_resource_process_list);
	g_test_add_func("/ConstrainedResource/process_list",
			test_constrained_resource_process_list);
	g_test_add_func("/ConstrainedResource/process",
			test_constrained_resource_process);
	g_test_add_func("/ResourcesTable/resources_table", test_resources_table);
	g_test_add_func("/Server/server_instantiation", test_server_instantiation);
	g_test_add_func("/Server/server_termination", test_server_termination);
	g_test_add_func("/Server/run", test_server_run);
	g_test_add_func("/Server/run_multiple_clients",
			test_server_run_multiple_clients);

	if (is_test_eeclust_node()) {
		// should only be executed on eeclust nodes...
		g_test_add_func("/Daemon/daemon_init", test_daemon_init);
		g_test_add_func("/Config/create_resources_table",
				test_config_create_resources_table);
#ifdef ENABLE_TRACE
		g_test_add_func("/Daemon/tracing", test_daemon_tracing);
#endif
	}
	// run tests from the suite
	gint ret = g_test_run();

	g_file_delete(g_file_new_for_path(config_file_path), NULL, NULL);

	g_free(config_file_path);

	return ret;
}
